import pandas as pd
import numpy as np
import copy


from chainladder.core.base import TriangleBase


class Triangle(TriangleBase):
    """
    The core data structure of the chainladder package

    Parameters
    ----------
    data : DataFrame
        A single dataframe that contains columns represeting all other
        arguments to the Triangle constructor
    origin : str or list
         A representation of the accident, reporting or more generally the
         origin period of the triangle that will map to the Origin dimension
    development : str or list
        A representation of the development/valuation periods of the triangle
        that will map to the Development dimension
    columns : str or list
        A representation of the numeric data of the triangle that will map to
        the columns dimension.  If None, then a single 'Total' key will be
        generated.
    index : str or list or None
        A representation of the index of the triangle that will map to the
        index dimension.  If None, then a single 'Total' key will be generated.
    origin_format : optional str
        A string representation of the date format of the origin arg. If
        omitted then date format will be inferred by pandas.
    development_format : optional str
        A string representation of the date format of the development arg. If
        omitted then date format will be inferred by pandas.
    cumulative : bool
        Whether the triangle is cumulative or incremental.  This attribute is
        required to use the `grain` and `dev_to_val` methods and will be
        automatically set when invoking `cum_to_incr` or `incr_to_cum` methods.

    Attributes
    ----------
    index : Series
        Represents all available levels of the index dimension.
    columns : Series
        Represents all available levels of the value dimension.
    origin : DatetimeIndex
        Represents all available levels of the origin dimension.
    development : Series
        Represents all available levels of the development dimension.
    valuation : DatetimeIndex
        Represents all valuation dates of each cell in the Triangle.
    shape : tuple
        The 4D shape of the triangle instance
    link_ratio, age_to_age
        Displays age-to-age ratios for the triangle.
    valuation_date : date
        The latest valuation date of the data
    loc : Triangle
        pandas-style ``loc`` accessor
    iloc : Triangle
        pandas-style ``iloc`` accessor
    latest_diagonal : Triangle
        The latest diagonal of the triangle
    is_cumulative: bool
        Whether the triangle is cumulative or not
    is_ultimate: bool
        Whether the triangle has an ultimate valuation
    is_val_tri:
        Whether the triangle development period is expressed as valuation
        periods.
    values : array
        4D numpy array underlying the Triangle instance
    T : Triangle
        Transpose index and columns of object.  Only available when Triangle is
        convertible to DataFrame.
    """
    @property
    def shape(self):
        return self.values.shape

    @property
    def index(self):
        return pd.DataFrame(list(self.kdims), columns=self.key_labels)

    @property
    def columns(self):
        return self._idx_table().columns

    @columns.setter
    def columns(self, value):
        self._len_check(self.columns, value)
        self.vdims = [value] if type(value) is str else value
        self._set_slicers()

    @property
    def origin(self):
        return pd.DatetimeIndex(self.odims, name='origin') \
                 .to_period(self.origin_grain)

    @origin.setter
    def origin(self, value):
        self._len_check(self.origin, value)
        value = pd.PeriodIndex([item for item in list(value)],
                               freq=self.origin_grain).to_timestamp()
        self.odims = value.values

    @property
    def development(self):
        return pd.Series(list(self.ddims), name='development').to_frame()

    @development.setter
    def development(self, value):
        self._len_check(self.development, value)
        self.ddims = [value] if type(value) is str else value

    @property
    def is_full(self):
        return self._nan_triangle().sum().sum() == np.prod(self.shape[-2:])

    @property
    def is_ultimate(self):
        return sum(self.valuation >= '2262') > 0

    @property
    def is_val_tri(self):
        return (type(self.ddims) != np.ndarray)

    @property
    def latest_diagonal(self):
        return self._get_latest_diagonal()

    @property
    def link_ratio(self):
        obj = copy.deepcopy(self)
        temp = obj.values.copy()
        temp[temp == 0] = np.nan
        val_array = obj.valuation.to_timestamp().values.reshape(
            obj.shape[-2:], order='f')[:, 1:]
        obj.values = temp[..., 1:]/temp[..., :-1]
        obj.ddims = np.array(['{}-{}'.format(obj.ddims[i], obj.ddims[i+1])
                              for i in range(len(obj.ddims)-1)])
        # Check whether we want to eliminate the last origin period
        if np.max(np.sum(~np.isnan(self.values[..., -1, :]), 2)-1) == 0:
            obj.values = obj.values[..., :-1, :]
            obj.odims = obj.odims[:-1]
            val_array = val_array[:-1, :]
        obj.valuation = pd.DatetimeIndex(
            pd.DataFrame(val_array).unstack().values).to_period(self._lowest_grain())
        return obj

    @property
    def age_to_age(self):
        return self.link_ratio

    # ---------------------------------------------------------------- #
    # ---------------------- End User Methods ------------------------ #
    # ---------------------------------------------------------------- #
    def _get_latest_diagonal(self, compress=True):
        ''' Method to return the latest diagonal of the triangle.  Requires
            self.nan_overide == False.

        Parameters
        ----------
        compress: bool
            Whether to collapse the diagonal into a single columns
        '''
        obj = copy.deepcopy(self)
        diagonal = obj[obj.valuation == obj.valuation_date].values
        if compress:
            diagonal = np.expand_dims(np.nansum(diagonal, 3), 3)
            obj.ddims = np.array([None])
            obj.valuation = pd.DatetimeIndex(
                [pd.to_datetime(obj.valuation_date)] *
                len(obj.odims)).to_period(self._lowest_grain())
        obj.values = diagonal
        return obj

    def incr_to_cum(self, inplace=False):
        """Method to convert an incremental triangle into a cumulative triangle.

        Parameters
        ----------
        inplace: bool
            Set to True will update the instance data attribute inplace

        Returns
        -------
            Updated instance of triangle accumulated along the origin
        """

        if inplace:
            if not self.is_cumulative:
                np.cumsum(np.nan_to_num(self.values), axis=3, out=self.values)
                self.values = self._expand_dims(self._nan_triangle())*self.values
                self.values[self.values == 0] = np.nan
                self.is_cumulative = True
            return self
        else:
            new_obj = copy.deepcopy(self)
            return new_obj.incr_to_cum(inplace=True)

    def cum_to_incr(self, inplace=False):
        """Method to convert an cumlative triangle into a incremental triangle.

        Parameters
        ----------
            inplace: bool
                Set to True will update the instance data attribute inplace

        Returns
        -------
            Updated instance of triangle accumulated along the origin
        """

        if inplace:
            if self.is_cumulative:
                temp = np.nan_to_num(self.values)[..., 1:] - \
                    np.nan_to_num(self.values)[..., :-1]
                temp = np.concatenate((self.values[..., 0:1], temp), axis=3)
                temp = temp*self._expand_dims(self._nan_triangle())
                temp[temp == 0] = np.nan
                self.values = temp
                self.is_cumulative = False
            return self
        else:
            new_obj = copy.deepcopy(self)
            return new_obj.cum_to_incr(inplace=True)

    def dev_to_val(self, inplace=False):
        ''' Converts triangle from a development lag triangle to a valuation
        triangle.

        Parameters
        ----------
        inplace : bool
            Whether to mutate the existing Triangle instance or return a new
            one.

        Returns
        -------
            Updated instance of triangle with valuation periods.
        '''
        if inplace:
            self = self._val_dev_chg('dev_to_val') if not self.is_val_tri else self
            return self
        return self._val_dev_chg('dev_to_val') if not self.is_val_tri else copy.deepcopy(self)

    def val_to_dev(self, inplace=False):
        ''' Converts triangle from a valuation triangle to a development lag
        triangle.

        Parameters
        ----------
        inplace : bool
            Whether to mutate the existing Triangle instance or return a new
            one.

        Returns
        -------
            Updated instance of triangle with development lags
        '''
        if not self.is_val_tri:
            ret_val = self
        else:
            if self.is_ultimate:
                obj = self[self.valuation<'2262']
                non_ultimates = obj.values
                obj = obj._val_dev_chg('val_to_dev')
                ultimate = self[self.valuation>='2262']
                max_val = self[self.valuation<'2262'][self.origin== self.origin.max()].valuation.max()
                max_dev = obj[obj.origin == obj.origin.max()]
                max_dev = max_dev[max_dev.valuation == max_val].ddims[0]
                obj = obj[obj.development <= max_dev]
                obj.values = np.concatenate(
                    (obj.values, ultimate.values), axis=-1)
                obj.ddims = np.append(obj.ddims, 9999)
                obj.valuation = obj._valuation_triangle(obj.ddims)
                obj.valuation_date = max(obj.valuation).to_timestamp()
                ret_val = obj
            else:
                ret_val = self._val_dev_chg('val_to_dev')
        ret_val.values = self._expand_dims(ret_val._nan_triangle())*ret_val.values
        if inplace:
            self = ret_val
        return ret_val


    def _val_dev_chg(self, kind):
        obj = copy.deepcopy(self)
        o_vals = obj._expand_dims(np.arange(len(obj.origin))[:, np.newaxis])
        if self.shape[-1] == 1:
            return obj
        if kind == 'val_to_dev':
            step = {'Y': 12, 'Q': 3, 'M': 1}[obj.development_grain]
            mtrx = \
                12*(obj.ddims.to_timestamp(how='e').year.values[np.newaxis] -
                    obj.origin.to_timestamp(how='s')
                       .year.values[:, np.newaxis]) + \
                   (obj.ddims.to_timestamp(how='e').month.values[np.newaxis] -
                    obj.origin.to_timestamp(how='s')
                       .month.values[:, np.newaxis]) + 1
            rng = range(mtrx[mtrx > 0].min(), mtrx.max()+1, step)
        else:
            rng = obj.valuation.unique().sort_values()
        old_arr = None
        for item in rng:
            if kind == 'val_to_dev':
                val = np.where(mtrx == item)
            else:
                val = np.where(obj._expand_dims(obj.valuation == item)
                                  .reshape(obj.shape, order='f'))[-2:]
            val = np.unique(np.array(list(zip(val[0], val[1]))), axis=0)
            arr = np.expand_dims(obj.values[:, :, val[:, 0], val[:, 1]], -1)
            if val[0, 0] != 0:
                prepend = obj._expand_dims(np.array([np.nan]*(val[0, 0]))[:, np.newaxis])
                arr = np.concatenate((prepend, arr), -2)
            if len(obj.origin)-1-val[-1, 0] != 0:
                append = obj._expand_dims(
                    np.array([np.nan]*(len(obj.origin)-1-val[-1, 0]))[:, np.newaxis])
                arr = np.concatenate((arr, append), -2)
            if obj.is_cumulative and old_arr is not None:
                arr = np.isnan(arr)*np.nan_to_num(old_arr) + np.nan_to_num(arr)
            old_arr = arr.copy()
            o_vals = np.append(o_vals, arr, -1)
        obj.values = o_vals[..., 1:]
        obj.values[obj.values == 0] = np.nan
        keep = np.where(np.sum(
            (np.nansum(np.nansum(obj.values, 0), 0)), -2) !=
            len(obj.origin))[-1]
        obj.values = obj.values[..., keep]
        if kind == 'val_to_dev':
            obj.ddims = np.array([item for item in rng])
            obj.ddims = obj.ddims[keep]
            obj.valuation = obj._valuation_triangle()
        else:
            obj.ddims = obj.valuation.unique().sort_values()
            obj.ddims = obj.ddims[keep]
            obj.values = obj.values[..., :np.where(
                obj.ddims.to_timestamp() <= obj.valuation_date)[0].max()+1]
            obj.ddims = obj.ddims[obj.ddims.to_timestamp()
                                  <= obj.valuation_date]
            obj.valuation = pd.PeriodIndex(
                np.repeat(obj.ddims.values[np.newaxis],
                          len(obj.origin)).reshape(1, -1).flatten())
        return obj


    def grain(self, grain='', inplace=False):
        """Changes the grain of a cumulative triangle.

        Parameters
        ----------
        grain : str
            The grain to which you want your triangle converted, specified as
            'O<x>D<y>' where <x> and <y> can take on values of ``['Y', 'Q', 'M'
            ]`` For example, 'OYDY' for Origin Year/Development Year, 'OQDM'
            for Origin quarter/Development Month, etc.
        incremental : bool
            Grain does not work on incremental triangles and this argument let's
            the function know to make it cumuative before operating on the grain.
        inplace : bool
            Whether to mutate the existing Triangle instance or return a new
            one.

        Returns
        -------
            Triangle
        """
        if not self.is_cumulative:
            # Must be cumulative to work
            obj = self.incr_to_cum().dev_to_val(inplace=True)
        else:
            obj = self.dev_to_val(inplace=True)
        # put data in valuation mode
        ograin_new = grain[1:2]
        ograin_old = obj.origin_grain
        if ograin_new != ograin_old:
            o_dt = pd.Series(obj.odims)
            if ograin_new == 'Q':
                o = np.array(pd.to_datetime(
                    o_dt.dt.year.astype(str) + 'Q' + o_dt.dt.quarter.astype(str)))
            elif ograin_new == 'Y':
                o = np.array(pd.to_datetime(o_dt.dt.year, format='%Y'))
            else:
                o = obj.odims
            o_new = np.unique(o)
            o = np.repeat(np.expand_dims(o, axis=1), len(o_new), axis=1)
            o_new = np.repeat(o_new[np.newaxis], len(o), axis=0)
            o_bool = np.repeat((o == o_new)[:, np.newaxis],
                               len(obj.ddims), axis=1)
            o_bool = obj._expand_dims(o_bool)
            new_tri = np.repeat(np.nan_to_num(obj.values)[..., np.newaxis],
                                o_bool.shape[-1], axis=-1)
            new_tri[~np.isfinite(new_tri)] = 0
            new_tri = np.swapaxes(np.sum(new_tri*o_bool, axis=2), -1, -2)
            obj.values = new_tri
            obj.odims = np.unique(o)
            obj.valuation = obj._valuation_triangle()
            del obj._nan_triangle_
        obj = obj.val_to_dev(inplace=True)
        # Now do development
        dev_grain_dict = {'M': {'Y': 12, 'Q': 3, 'M': 1},
                          'Q': {'Y': 4, 'Q': 1},
                          'Y': {'Y': 1}}
        dgrain_new = grain[-1]
        dgrain_old = obj.development_grain
        if obj.shape[3] != 1:
            keeps = dev_grain_dict[dgrain_old][dgrain_new]
            keeps = np.where(np.arange(obj.shape[3]) % keeps == 0)[0]
            keeps = -(keeps + 1)[::-1]
            obj.values = obj.values[..., keeps]
            obj.ddims = obj.ddims[keeps]
        obj.origin_grain = ograin_new
        obj.development_grain = dgrain_new
        obj.values[obj.values == 0] = np.nan
        obj.valuation = obj._valuation_triangle()
        if hasattr(obj, '_nan_triangle_'):
            # Force update on _nan_triangle at next access.
            del obj._nan_triangle_
        if not self.is_cumulative:
            obj = obj.cum_to_incr()
        if inplace:
            self = obj
            return self
        else:
            return obj


    def trend(self, trend=0.0, axis='origin'):
        """  Allows for the trending of a Triangle object or an origin vector.
        This method trends using days and assumes a years is 365.25 days long.

        Parameters
        ----------
        trend : float
            The annual amount of the trend
        axis : str (options: ['origin', 'valuation'])
            The axis on which to apply the trend

        Returns
        -------
        Triangle
            updated with multiplicative trend applied.
        """
        days = np.datetime64(self.valuation_date)
        if axis == 'origin':
            trend = np.array((1 + trend)**-(
                pd.Series(self.origin.end_time.values-days).dt.days/365.25)
                )[np.newaxis, np.newaxis, ..., np.newaxis]
        elif axis == 'valuation':
            trend = (1 + trend)**-(
                pd.Series(self.valuation.end_time.values-days)
                .dt.days.values.reshape(self.shape[-2:], order='f')/365.25)
        obj = copy.deepcopy(self)
        obj.values = obj.values*trend
        return obj



import pandas as pd
import copy
import json


class _Workbook:
    """
    Excel Workbook level configurations.  This is not part of the end_user API
    """
    font_name = 'Calibri'
    font_size = 11
    max_column_width = 30
    max_portrait_width = 120
    footer = '&CPage &P of &N\n&A'

    def __init__(self, workbook_path, exhibits):
        title_format = \
            [{'font_size': 20, 'align': 'center', 'font_name': self.font_name},
            {'font_size': 16, 'align': 'center', 'font_name': self.font_name},
            {'font_size': 16, 'align': 'center', 'font_name': self.font_name},
            {'font_size': 13, 'align': 'center', 'font_name': self.font_name}]
        index_format = {'num_format': '0;(0)', 'text_wrap': True,
                        'bold': True, 'valign': 'bottom', 'align': 'right',
                        'font_name': self.font_name}
        header_format = {'num_format': '0;(0)', 'text_wrap': True, 'bottom': 1,
                        'bold': True, 'valign': 'bottom', 'align': 'left',
                        'font_name': self.font_name}
        all_columns = {'align': 'center', 'font_name': self.font_name}
        decimal_format = {'num_format': '#,0.00', 'font_name': self.font_name}
        datetime_format = {'num_format': 'yyyy-mm-dd hh:mm', 'font_name': self.font_name}
        date_format = {'num_format': 'yyyy-mm-dd', 'font_name': self.font_name}
        int_format = {'num_format': '#,0', 'font_name': self.font_name}
        text_format = {'align': 'left', 'font_name': self.font_name}
        self.formats = {}
        self.writer = pd.ExcelWriter(
            workbook_path)
        self.title_format = [self.writer.book.add_format(item)
                        for item in title_format]
        self.header_format = self.writer.book.add_format(header_format)
        self.index_format = self.writer.book.add_format(index_format)
        self.global_formats = {
            'font_name': self.font_name,
            'font_size': self.font_size}
        self.default_formats = {
            '"float64"': self.writer.book.add_format(decimal_format),
            '"float32"': self.writer.book.add_format(decimal_format),
            '"int64"': self.writer.book.add_format(int_format),
            '"int32"': self.writer.book.add_format(int_format),
            '"<M8[ns]"': self.writer.book.add_format(datetime_format),
            '"object"': self.writer.book.add_format(text_format),
        }
        self.exhibits = exhibits
        self.workbook_path = workbook_path

    def __repr__(self):
        return self.workbook_path

    def to_excel(self):
        """ Outputs object to Excel.

        Parameters:
        -----------
        workbook_path : str
            The target path and filename of the Excel document
        """
        #all_formats = {}
        #for item in self.exhibits:
        #    all_formats.update(item.formats)
        if self.exhibits.__class__.__name__ != 'Tabs':
            self.exhibits = Tabs(('sheet1', self.exhibits))
        for sheet in self.exhibits:
            self._write(sheet[1], sheet[0])
        self.writer.save()
        self.writer.close()

    def _write(self, exhibit, sheet, start_row=0, start_col=0):
        klass = exhibit.__class__.__name__
        if getattr(exhibit, 'title', None) is not None:
            t = copy.deepcopy(exhibit.title)
            exhibit.title = None
            exhibit = Column(t, exhibit)
            self._write(exhibit, sheet, start_row, start_col)
        elif klass in ['Row', 'Column']:
            start_row = start_row + exhibit.margin[0]
            start_col = start_col + exhibit.margin[3]
            for item in exhibit.args:
                self._write(item, sheet, start_row, start_col)
                if klass == 'Column':
                    start_row = start_row + item.height
                if klass == 'Row':
                   start_col = start_col + item.width
        else:
            exhibit.start_row = start_row
            exhibit.start_col = start_col
            exhibit.sheet_name = sheet
            try:
                exhibit.worksheet = self.writer.sheets[exhibit.sheet_name]
            except:
                pd.DataFrame().to_excel(self.writer, sheet_name=exhibit.sheet_name)
                exhibit.worksheet = self.writer.sheets[exhibit.sheet_name]
            if klass == 'DataFrame':
                if exhibit.header:
                    self._write_header(exhibit)
                if exhibit.index:
                    self._write_index(exhibit)
                self._register_formats(exhibit)
                self._write_data(exhibit)
            if klass == '_Title':
                self._write_title(exhibit)
            #self._set_worksheet_properties(exhibit)

    def set_worksheet_properties(self, exhibit):
        ''' Format column widths, headers footers, etc.'''
        # exhibit.worksheet.hide_gridlines(2)
        exhibit.worksheet.set_footer(self.footer)
        widths = [min(self.max_column_width, item)
                  for item in exhibit.column_widths]
        widths[0] = 18 if widths[0] < 18 else widths[0]
        for num, item in enumerate(widths):
            col = exhibit.start_col + exhibit.index + num
            exhibit.worksheet.set_column(col, col, item)
        if sum(widths) > self.max_portrait_width:
            exhibit.worksheet.set_landscape()

    def _write_title(self, exhibit):
        start_row = exhibit.start_row
        start_col = exhibit.start_col
        end_row = start_row + exhibit.height
        end_col = start_col + exhibit.width - 1
        row_rng = range(start_row, end_row)
        for r in row_rng:
            exhibit.worksheet.merge_range(
                r, start_col, r, end_col, exhibit.data.iloc[r - start_row][0],
                self.title_format[r - start_row])

    def _write_header(self, exhibit):
        ''' Adds column headers to data table '''
        if not exhibit.index:
            headers = exhibit.data.columns
        else:
            headers = [exhibit.index_label]+list(exhibit.data.columns)
        for col_num, value in enumerate(headers):
            exhibit.worksheet.write(
                exhibit.start_row + exhibit.margin[0],
                col_num + exhibit.start_col + exhibit.margin[3],
                value, self.header_format)
            if exhibit.col_nums:
                exhibit.worksheet.write(
                    exhibit.start_row + 1,
                    col_num, -col_num-1, self.header_format)

    def _write_index(self, exhibit):
        ''' Adds row index to data table '''
        for row_num, value in enumerate(exhibit.data.index.astype(str)):
            exhibit.worksheet.write(
                row_num + exhibit.start_row + exhibit.header + \
                exhibit.col_nums + exhibit.margin[0],
                exhibit.start_col + exhibit.margin[3],
                value, self.index_format)

    def _register_formats(self, exhibit):
        """
        Registers all unique user-defined formats with the Workbook
        """
        for num, k in enumerate(exhibit.formats.keys()):
            v = exhibit.formats[k]
            if type(v) is dict:
                col_formats = v
            elif self.default_formats.get(v, None) is not None:
                col_formats = self.default_formats[v]
            elif type(v) is str:
                col_formats = {'num_format': v}
            else:
                raise ValueError(f'Cannot infer format {v}')
            for desc, attr in self.global_formats.items():
                if col_formats is None:
                    col_formats = attr
            if self.formats.get(json.dumps(col_formats), None) is None:
                self.formats[json.dumps(col_formats)] = \
                    self.writer.book.add_format(col_formats)
        for k, v in exhibit.formats.items():
            exhibit.formats[k] = self.formats.get(json.dumps(v),
                                self.default_formats.get(json.dumps(v)))

    def _write_data(self, exhibit):
        start_row = exhibit.start_row + exhibit.col_nums + exhibit.header + \
                    exhibit.margin[0]
        start_col = exhibit.start_col + exhibit.index + exhibit.margin[3]
        end_row = start_row + exhibit.data.shape[0]
        end_col = start_col + exhibit.data.shape[1]
        row_rng = range(start_row, end_row)
        col_rng = range(start_col, end_col)
        d = exhibit.data.fillna('').values
        for c in col_rng:
            c_idx = c - exhibit.index - exhibit.start_col - exhibit.margin[3]
            fmt = exhibit.formats[exhibit.data.columns[c_idx]]
            for r in row_rng:
                r_idx = r - exhibit.col_nums - exhibit.header - \
                        exhibit.start_row - exhibit.margin[0]
                exhibit.worksheet.write(r, c, d[r_idx, c_idx], fmt)
                if r == start_row:
                    exhibit.worksheet.set_column(
                        first_col=c, last_col=c,
                        width=exhibit.column_widths[c_idx])


class _Title:
    def __init__(self, data, width):
        if type(data) is str:
            data = [data]
        self.data = pd.DataFrame(data)
        self.width = width
        self.height = len(self.data)
        self.header=False
        self.index=False
        self.col_nums=False
        self.margin=(0,0,0,0)
        self.formats = {}

    def __len__(self):
        return len(self.data)

class DataFrame:
    """
    Excel-ready DataFrame

    Parameters:
    -----------
    data : DataFrame or Triangle (2D)
        The data to be places in the exhibit
    header : bool or list (len of data.columns)
        False uses no headers, True uses headers from data. Alternatively,
        a list of strings will override headers.
    formats : str or list (len of data.columns)
        The formats to be applied to the data.  Options include
        'money', 'percent', 'decimal', 'date', 'int', and 'text'.  Each
        format can be overriden at the class level by overriding its:
        resepective format dict (e.g. DataFrame.money_format, ...)
    index : bool, default True
        Write row names (index).
    index_label : str or sequence, optional
        Column label for index column(s) if desired.
    title : list
        A list of strings up to length 4 (Title, subtitle1, subtitle2,
        subtitle3) to be placed above the data in the exhibit.
    col_nums : bool
        Set to True will insert column numbers into the exhibit.
    """

    min_numeric_col_width = 12
    # Padding since bold characters are slightly larger than regular
    # and need a bit more width
    col_padding_multiplier = 1.1

    def __init__(self, data, header=True, formats=None, index=True,
                 index_label='', title=None, col_nums=False,
                 column_widths=None, margin=None):
        if type(data) is not pd.DataFrame:
            data = data.to_frame()
        self.margin = _Container._set_margin(margin)
        self.data = data
        self.header = header
        self.index = index
        self.index_label = index_label
        self.col_nums = col_nums
        self.format_validation(formats)
        if column_widths is None:
            self.column_widths = self.get_column_widths()
        else:
            self.column_widths = column_widths
        self.height = data.shape[0] + self.col_nums + self.header + \
                      self.margin[0] + self.margin[2]
        self.width = data.shape[1] + self.index + self.margin[1] + \
                     self.margin[3]
        if title is None or title == []:
            title = None
        elif len(title) < 4:
            self.height = self.height + len(title)
            title = _Title(title, width=self.width - self.margin[1] - self.margin[3])
        self.title = title

    def to_excel(self, workbook_path):
        """ Outputs object to Excel.

        Parameters:
        -----------
        workbook_path : str
            The target path and filename of the Excel document
        """
        _Workbook(workbook_path=workbook_path, exhibits=self).to_excel()

    def get_column_widths(self):
        header_w = [max([len(token) for token in str(item).split(' ')])
                    * self.col_padding_multiplier
                    for item in self.data.columns]
        numeric_cols = self.data.select_dtypes('number').columns
        row_w = [(self.min_numeric_col_width if item in numeric_cols
                  else max(self.data[item].astype(str).str.len())
                  * self.col_padding_multiplier)
                 for item in self.data.columns]
        return [max(item) for item in zip(header_w, row_w)]

    def __repr__(self):
        return str(self.data.shape)

    def format_validation(self, formats):
        ''' Creates an Excel format compatible dictionary '''
        self.formats = dict(self.data.dtypes.astype(str))
        if type(formats) is list:
            self.formats.update(dict(zip(self.data.columns, formats)))
        elif type(formats) is str:
            self.formats.update(dict(zip(
                self.data.columns,
                [{'num_format': formats}] * len(self.data.columns))))
        elif type(formats) is dict:
            if list(formats.keys())[0] not in self.data.columns:
                self.formats.update(dict(zip(
                    self.data.columns,
                    [formats] * len(self.data.columns))))
            else:
                self.formats.update(formats)
        else:
            pass
        for field, format_str in self.formats.items():
            digits = 0
            if type(format_str) is not str:
                continue
            if format_str.find('[') > 0 and \
               format_str not in ('datetime64[ns]', 'timedelta64[ns]'):
                digits = int(
                    format_str[format_str.find('[') + 1:format_str.find(']')])
            digits = ('.' if digits > 0 else '') + ''.join(['0']*digits)
            if format_str.lower().find('currency') > -1:
                self.formats[field] = (
                    f'_($* #,##0{digits}_);'
                    f'_($* (#,##0{digits});'
                    '_($* "-"??_);'
                    '_(@_)')
            if format_str.lower().find('accounting') > -1:
                self.formats[field] = (
                    f'_(* #,##0{digits}_);'
                    f'_(* (#,##0{digits});'
                    '_(* "-"??_);'
                    '_(@_)')
            if format_str.lower().find('percent') > -1:
                self.formats[field] = f'0{digits}%'


class _Container():
    """ Base class for Row and Column
    """
    def __init__(self, *args, **kwargs):
        if len(args) != len(set(args)):
            self.args = tuple([copy.deepcopy(item) for item in args])
        else:
            self.args = args
        self.margin = _Container._set_margin(kwargs.get('margin', None))
        self._title_len = 0
        if kwargs.get('title', None) is not None:
            self.title = _Title(kwargs['title'], width=self.width)
            if getattr(self, 'title', None) is not None:
                self._title_len = len(getattr(self, 'title'))
        arg_title = [0]
        for item in args:
            if getattr(item, 'title', None) is not None:
                arg_title.append(len(getattr(item, 'title')))
        self._arg_title_len = arg_title

    def __getitem__(self, key):
        return self.args[key]

    def __len__(self):
        return len(self.args)

    def to_excel(self, workbook_path):
        """ Outputs object to Excel.

        Parameters:
        -----------
        workbook_path : str
            The target path and filename of the Excel document
        """
        _Workbook(workbook_path=workbook_path, exhibits=self).to_excel()

    @staticmethod
    def _set_margin(margin):
        if type(margin) is int:
            return (margin, margin, margin, margin)
        elif margin is None:
            return (0, 0, 0, 0)
        elif type(margin) is tuple and len(margin) == 2:
            return (margin[0], margin[1], margin[0], margin[1])
        else:
            return margin


class Row(_Container):
    """
    Lay out child components in a single horizontal row.
    Children can be specified as positional arguments, as a single argument
    that is a sequence.

    Parameters
    ----------
    args:
        Children can be of the chainlader DataFrame, Row, and Column classes.
    title: optional (str or list)
        The title to be displayed across the top of the container.  Must be
        specified using the keyword `title=`
    margin:
        Cell margins to put around the container.  Can be expressed as an int
        or a 2-tuple or a 4-tuple.  The tuple style follows CSS margin guidelines
        which generally start from the top and work around clockwise.

    Attributes
    ----------
    height : int
        Height of the container and is a function of the elements it contains
    width : int
        Width of the container and is a function of the elements it contains

    """
    @property
    def height(self):
        margin = self.margin[0] + self.margin[2]
        return max([item.height for item in self.args]) + \
               max(self._arg_title_len) + margin + self._title_len

    @property
    def width(self):
        margin = self.margin[1] + self.margin[3]
        return sum([item.width for item in self.args]) + margin


class Column(_Container):
    """
    Lay out child components in a single vertical column.
    Children can be specified as positional arguments, as a single argument
    that is a sequence.

    Parameters
    ----------
    args:
        Children can be of the chainlader DataFrame, Row, and Column classes.
    title: optional (str or list)
        The title to be displayed across the top of the container.  Must be
        specified using the keyword `title=`
    margin:
        Cell margins to put around the container.  Can be expressed as an int
        or a 2-tuple or a 4-tuple.  The tuple style follows CSS margin guidelines
        which generally start from the top and work around clockwise.

    Attributes
    ----------
    height : int
        Height of the container and is a function of the elements it contains
    width : int
        Width of the container and is a function of the elements it contains

    """
    @property
    def height(self):
        margin = self.margin[0] + self.margin[2]
        return sum([item.height for item in self.args]) + \
               sum(self._arg_title_len) + margin + self._title_len

    @property
    def width(self):
        margin = self.margin[1] + self.margin[3]
        return max([item.width for item in self.args]) + margin


class Tabs:
    """
    Layout exhibits across worksheets.

    Parameters
    ----------
    args:
        Children must be a tuple with a sheet name and any of chainlader
        DataFrame, Row, and Column classes.  For example,
        ('sheet1', cl.DataFrame(data))
    """

    def __init__(self, *args, **kwargs):
        if len(args) != set([item[1] for item in args]):
            self.args = tuple([(item[0], copy.deepcopy(item[1]))
                               for item in args])
        else:
            self.args = args

    def __getitem__(self, key):
        return self.args[key]

    def __len__(self):
        return len(self.args)

    def to_excel(self, workbook_path):
        """ Outputs object to Excel.

        Parameters:
        -----------
        workbook_path : str
            The target path and filename of the Excel document
        
        """
        _Workbook(workbook_path=workbook_path, exhibits=self).to_excel()



import pandas as pd
import numpy as np
import json
import joblib


class TriangleIO():
    def to_pickle(self, path, protocol=None):
        joblib.dump(self, filename=path, protocol=protocol)

    def to_json(self):
        ''' Serializes triangle object to json format

        Returns
        -------
            string representation of object in json format
        '''
        json_dict = {}
        attributes = ['values', 'kdims', 'vdims', 'odims', 'ddims']
        for attribute in attributes:
            json_dict[attribute] = {
                'dtype': str(getattr(self, attribute).dtype),
                'array': getattr(self, attribute).tolist()}
        json_dict['key_labels'] = self.key_labels
        json_dict['origin_grain'] = self.origin_grain
        json_dict['development_grain'] = self.development_grain
        json_dict['nan_override'] = self.nan_override
        json_dict['valuation_date'] = self.valuation_date.strftime('%Y-%m-%d')
        return json.dumps(json_dict)


class EstimatorIO:
    ''' Class intended to allow persistence of estimator objects
    '''

    def to_pickle(self, path, protocol=None):
        joblib.dump(self, filename=path, protocol=protocol)

    def to_json(self):
        ''' Serializes triangle object to json format

        Returns
        -------
            string representation of object in json format
        '''
        return json.dumps(
            {'params': self.get_params(),
             '__class__': self.__class__.__name__})

    def __contains__(self, value):
        if self.__dict__.get(value, None) is None:
            return False
        return True



import numpy as np
from sklearn.base import BaseEstimator

class WeightedRegression(BaseEstimator):
    ''' Helper class that fits a system of regression equations
        simultaneously on a multi-dimensional array.  Look into
        SUR as a replacement.
    '''
    def __init__(self, axis=None, thru_orig=False):
        self.axis = axis
        self.thru_orig = thru_orig

    def infer_x_w(self):
        if self.w is None:
            self.w = np.ones(self.y.shape)
        if self.x is None:
            self.x = np.cumsum(np.ones(self.y.shape), self.axis)
        return self

    def fit(self, X, y=None, sample_weight=None):
        self.x = X
        self.y = y
        self.w = sample_weight
        if self.x is None:
            self.infer_x_w()
        if self.thru_orig:
            self._fit_OLS_thru_orig()
        else:
            self._fit_OLS()
        return self

    def _fit_OLS(self):
        ''' Given a set of w, x, y, and an axis, this Function
            returns OLS slope and intercept.
            TODO:
                Make this work with n_periods = 1 without numpy warning.
        '''
        w, x, y, axis = self.w.copy(), self.x.copy(), self.y.copy(), self.axis
        x[w == 0] = np.nan
        y[w == 0] = np.nan
        slope = (
            (np.nansum(w*x*y, axis)-np.nansum(x*w, axis)*np.nanmean(y, axis)) /
            (np.nansum(w*x*x, axis)-np.nanmean(x, axis)*np.nansum(w*x, axis)))
        intercept = np.nanmean(y, axis) - slope * np.nanmean(x, axis)
        self.slope_ = np.expand_dims(slope, -1)
        self.intercept_ = np.expand_dims(intercept, -1)
        return self

    def _fit_OLS_thru_orig(self):
        w, x, y, axis = self.w, self.x, self.y, self.axis
        coef = np.nansum(w*x*y, axis)/np.nansum((y*0+1)*w*x*x, axis)
        fitted_value = np.repeat(np.expand_dims(coef, axis),
                                 x.shape[axis], axis)
        fitted_value = (fitted_value*x*(y*0+1))
        residual = (y-fitted_value)*np.sqrt(w)
        wss_residual = np.nansum(residual**2, axis)
        mse_denom = np.nansum(y*0+1, axis)-1
        mse_denom[mse_denom == 0] = np.nan
        mse = wss_residual / mse_denom
        std_err = np.sqrt(mse/np.nansum(w*x*x*(y*0+1), axis))
        std_err = np.expand_dims(std_err, -1)
        std_err[std_err == 0] = np.nan
        coef = np.expand_dims(coef, -1)
        sigma = np.expand_dims(np.sqrt(mse), -1)
        self.slope_ = coef
        self.sigma_ = sigma
        self.std_err_ = std_err
        return self

    def sigma_fill(self, interpolation):
        ''' This Function is designed to take an array of sigmas and does log-
            linear extrapolation where n_obs=1 and sigma cannot be calculated.
        '''
        if interpolation == 'log-linear':
            self.sigma_ = self.loglinear_interpolation(self.sigma_)
        if interpolation == 'mack':
            self.sigma_ = self.mack_interpolation(self.sigma_)
        return self

    def std_err_fill(self):
        '''currently handled in development.py which doesn't feel right'''
        return self

    def loglinear_interpolation(self, y):
        ''' Use Cases: generally for filling in last element of sigma_
        '''
        y[y == 0] = np.nan
        ly = np.log(y)
        w = np.nan_to_num(ly*0+1)
        reg = WeightedRegression(self.axis, False).fit(None, ly, w)
        slope, intercept = reg.slope_, reg.intercept_
        fill_ = np.exp(reg.x*slope+intercept)*(1-w)
        return np.nan_to_num(y) + fill_

    def mack_interpolation(self, y):
        ''' Use Mack's approximation to fill last element of sigma_ which is the
            same as loglinear extrapolation using the preceding two element to
            the missing value. This function needs a recursive definition...
        '''
        w = np.nan_to_num(y*0+1)
        slicer_n, slicer_d, slicer_a = \
            ([slice(None)]*4), ([slice(None)]*4), ([slice(None)]*4)
        slicer_n[self.axis], slicer_d[self.axis], slicer_a[self.axis] = \
            slice(1, -1, 1), slice(0, -2, 1), slice(0, 2, 1)
        slicer_n, slicer_d, slicer_a = (tuple(slicer_n), tuple(slicer_d),
                                        tuple(slicer_a))
        fill_ = np.sqrt(
            abs(np.minimum((y[slicer_n]**4 / y[slicer_d]**2),
                np.minimum(y[slicer_d]**2, y[slicer_n]**2))))
        fill_ = np.concatenate((w[slicer_a], fill_), axis=self.axis)*(1-w)
        return np.nan_to_num(y) + fill_



from chainladder.development.base import Development, DevelopmentBase # noqa (API import)
from chainladder.development.munich import MunichAdjustment # noqa (API import)
from chainladder.development.incremental import IncrementalAdditive # noqa (API import)
from chainladder.development.bootstrap import BootstrapODPSample # noqa (API import)
from chainladder.development.constant import DevelopmentConstant # noqa (API import)


import pytest
from numpy.testing import assert_allclose
import chainladder as cl


@pytest.fixture
def atol():
    return 1e-5


data = ['RAA', 'ABC', 'GenIns', 'MW2008', 'MW2014']


@pytest.mark.parametrize('data', data)
def test_benktander_to_chainladder(data, atol):
    tri = cl.load_dataset(data)
    a = cl.Chainladder().fit(tri).ibnr_
    b = cl.Benktander(apriori=.8, n_iters=255).fit(tri, sample_weight=a).ibnr_
    assert_allclose(a.values, b.values, atol=atol)


def test_bf_eq_cl_when_using_cl_apriori():
    cl_ult = cl.Chainladder().fit(cl.load_dataset('quarterly')).ultimate_
    cl_ult.rename('development', ['apriori'])
    bf_ult = cl.BornhuetterFerguson().fit(cl.load_dataset('quarterly'),
                                          sample_weight=cl_ult).ultimate_
    assert_allclose(cl_ult.values, bf_ult.values, atol=1e-5)



from chainladder.tails import TailBase
from chainladder.utils import WeightedRegression
from chainladder.development import DevelopmentBase, Development
import numpy as np


class TailCurve(TailBase):
    """Allows for extraploation of LDFs to form a tail factor.

    Parameters
    ----------
    curve : str ('exponential', 'inverse_power')
        The type of curve extrapolation you'd like to use
    fit_period : slice
        A slice object representing the range (by index) of ldfs to use in
        the curve fit.
    extrap_periods : int
        Then number of development periods from attachment point to extrapolate
        the fit.
    errors : str ('raise' or 'ignore')
        Whether to raise an error or ignore observations that violate the
        distribution being fit.  The most common is ldfs < 1.0 will not work
        in either the ``exponential`` or ``inverse_power`` fits.

    Attributes
    ----------
    ldf_ :
        ldf with tail applied.
    cdf_ :
        cdf with tail applied.
    sigma_ :
        sigma with tail factor applied.
    std_err_ :
        std_err with tail factor applied
    """
    def __init__(self, curve='exponential', fit_period=slice(None, None, None),
                 extrap_periods=100, errors='ignore'):
        self.curve = curve
        self.fit_period = fit_period
        self.extrap_periods = extrap_periods
        self.errors = errors

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the tail will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        super().fit(X, y, sample_weight)
        _y = self.ldf_.values[..., :-1].copy()
        _w = np.zeros(_y.shape)
        if type(self.fit_period) is not slice:
            raise TypeError('fit_period must be slice.')
        else:
            _w[..., self.fit_period] = 1.0
        if self.errors == 'ignore':
            _w[_y <= 1.0] = 0
            _y[_y <= 1.0] = 1.01
        elif self.errors == 'raise' and np.any(y < 1.0):
            raise ZeroDivisionError('Tail fit requires all LDFs to be' +
                                    ' greater than 1.0')
        _y = np.log(_y - 1)
        n_obs = X.shape[-1]-1
        k, v = X.shape[:2]
        _x = self._get_x(_w, _y)
        # Get LDFs
        coefs = WeightedRegression(axis=3).fit(_x, _y, _w)
        slope, intercept = coefs.slope_, coefs.intercept_
        extrapolate = np.cumsum(
            np.ones(tuple(list(_y.shape)[:-1] +
                    [self.extrap_periods])), -1) + n_obs
        tail = self._predict_tail(slope, intercept, extrapolate)
        self.ldf_.values = self.ldf_.values[..., :-tail.shape[-1]]
        self.ldf_.values = np.concatenate((self.ldf_.values, tail), -1)
        obj = Development().fit_transform(X) if 'ldf_' not in X else X
        sigma, std_err = self._get_tail_stats(obj)
        self.sigma_.values[..., -1] = sigma[..., -1]
        self.std_err_.values[..., -1] = std_err[..., -1]
        self.slope_ = slope
        self.intercept_ = intercept
        self.cdf_ = DevelopmentBase._get_cdf(self)
        return self

    def _get_tail_weighted_time_period(self, X):
        """ Method to approximate the weighted-average development age of tail
        using log-linear extrapolation

        Returns: float32
        """
        y = X.ldf_.values.copy()
        y[y <= 1] = np.nan
        reg = WeightedRegression(axis=3).fit(None, np.log(y - 1), None)
        tail = np.prod(self.ldf_.values[..., -self._ave_period[0]-1:],
                       -1, keepdims=True)
        reg = WeightedRegression(axis=3).fit(None, np.log(y - 1), None)
        time_pd = (np.log(tail-1)-reg.intercept_)/reg.slope_
        return time_pd

    def _get_tail_stats(self, X):
        """ Method to approximate the tail sigma using
        log-linear extrapolation applied to tail average period
        """
        time_pd = self._get_tail_weighted_time_period(X)
        reg = WeightedRegression(axis=3).fit(None, np.log(X.sigma_.values), None)
        sigma_ = np.exp(time_pd*reg.slope_+reg.intercept_)
        y = X.std_err_.values
        y[y == 0] = np.nan
        reg = WeightedRegression(axis=3).fit(None, np.log(y), None)
        std_err_ = np.exp(time_pd*reg.slope_+reg.intercept_)
        return sigma_, std_err_

    def _get_x(self, w, y):
        # For Exponential decay, no transformation on x is needed
        if self.curve == 'exponential':
            return None
        if self.curve == 'inverse_power':
            reg = WeightedRegression(3, False).fit(None, y, w).infer_x_w()
            return np.log(reg.x)

    def _predict_tail(self, slope, intercept, extrapolate):
        if self.curve == 'exponential':
            tail_ldf = np.exp(slope*extrapolate + intercept)
        if self.curve == 'inverse_power':
            tail_ldf = np.exp(intercept)*(extrapolate**slope)
        return self._get_tail_prediction(tail_ldf)



import numpy as np
import pandas as pd
import chainladder as cl


def test_simple_exhibit():
    raa = cl.load_dataset('raa')

    col = cl.Column(
        cl.DataFrame(raa, margin=(0,0,1,0)),
        cl.DataFrame(raa.link_ratio, formats={'italic': True})
    )
    composite = cl.Row(
        col, col,
        title=['This title spans both Column Objects'],
        margin=(0,1,0,0)
    )
    x = cl.Tabs(
       ('a_sheet', composite),
       ('another_sheet', composite)
    ).to_excel('workbook.xlsx')



import numpy as np
import pandas as pd
import copy
from chainladder.methods import Chainladder


class MackChainladder(Chainladder):
    """ Basic stochastic chainladder method popularized by Thomas Mack

    Parameters
    ----------
    None

    Attributes
    ----------
    triangle
        returns **X**
    ultimate_
        The ultimate losses per the method
    ibnr_
        The IBNR per the method
    full_expectation_
        The ultimates back-filled to each development period in **X** replacing
        the known data
    full_triangle_
        The ultimates back-filled to each development period in **X** retaining
        the known data
    summary_
        summary of the model
    full_std_err_
        The full standard error
    total_process_risk_
        The total process error
    total_parameter_risk_
        The total parameter error
    mack_std_err_
        The total prediction error by origin period
    total_mack_std_err_
        The total prediction error across all origin periods

    """

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Data to which the model will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        super().fit(X, y, sample_weight)
        self._mack_recursion('param_risk')
        self._mack_recursion('process_risk')
        self._mack_recursion('total_param_risk')
        return self

    @property
    def full_std_err_(self):
        obj = copy.copy(self.X_)
        tri_array = self.full_triangle_.values
        weight_dict = {'regression': 0, 'volume': 1, 'simple': 2}
        val = np.array([weight_dict.get(item.lower(), 2)
                        for item in list(self.average_) + ['volume']])
        val = np.broadcast_to(val, self.X_.shape)
        weight = np.sqrt(tri_array[..., :len(self.X_.ddims)]**(2-val))
        weight[weight == 0] = np.nan
        obj.values = self.X_.sigma_.values / weight
        w = np.concatenate(
            (self.X_.w_, np.ones((*val.shape[:3], 1))*np.nan), axis=3)
        w[np.isnan(w)] = 1
        obj.values = np.nan_to_num(obj.values) * w
        obj.nan_override = True
        obj._set_slicers()
        return obj

    @property
    def total_process_risk_(self):
        origin = 2
        obj = copy.copy(self.process_risk_)
        obj.values = np.sqrt(np.nansum(self.process_risk_.values**2, origin))
        obj.values = np.expand_dims(obj.values, origin)
        obj.odims = ['tot_proc_risk']
        obj._set_slicers()
        return obj

    def _mack_recursion(self, est):
        obj = copy.copy(self.X_)
        nans = self.X_._nan_triangle()[np.newaxis, np.newaxis]
        nans = nans * np.ones(self.X_.shape)
        nans = np.concatenate(
            (nans, np.ones((*self.X_.shape[:3], 1))*np.nan), 3)
        nans = 1-np.nan_to_num(nans)
        properties = self.full_triangle_
        obj.valuation = properties.valuation
        obj.ddims = np.append(
            properties.ddims[:len(self.X_.ddims)],
            properties.ddims[-1])
        obj.nan_override = True
        risk_arr = np.zeros((*self.X_.shape[:3], 1))
        if est == 'param_risk':
            obj.values = self._get_risk(nans, risk_arr,
                                        obj.std_err_.values)
            obj._set_slicers()
            self.parameter_risk_ = obj
        elif est == 'process_risk':
            obj.values = self._get_risk(nans, risk_arr,
                                        self.full_std_err_.values)
            obj._set_slicers()
            self.process_risk_ = obj
        else:
            risk_arr = risk_arr[..., 0:1, :]
            obj.values = self._get_tot_param_risk(risk_arr)
            obj.odims = ['Total param risk']
            obj._set_slicers()
            self.total_parameter_risk_ = obj

    def _get_risk(self, nans, risk_arr, std_err):
        full_tri = self.full_triangle_.values[..., :len(self.X_.ddims)]
        t1_t = (full_tri * std_err)**2
        extend = self.X_.ldf_.shape[-1]-self.X_.shape[-1]+1
        ldf = self.X_.ldf_.values[..., :len(self.X_.ddims)-1]
        ldf = np.concatenate(
            (ldf, np.prod(self.X_.ldf_.values[..., -extend:], -1,
             keepdims=True)), -1)
        for i in range(len(self.X_.ddims)):
            t1 = t1_t[..., i:i+1]
            t2 = (ldf[..., i:i+1] * risk_arr[..., i:i+1])**2
            t_tot = np.sqrt(t1+t2)*nans[..., i+1:i+2]
            risk_arr = np.concatenate((risk_arr, t_tot), 3)
        return risk_arr

    def _get_tot_param_risk(self, risk_arr):
        """ This assumes triangle symmertry """
        t1 = self.full_triangle_.values[..., :len(self.X_.ddims)] - \
            np.nan_to_num(self.X_.values) + \
            np.nan_to_num(self.X_._get_latest_diagonal(False).values)
        t1 = np.sum(t1*self.X_.std_err_.values, axis=2, keepdims=True)
        extend = self.X_.ldf_.shape[-1]-self.X_.shape[-1]+1
        ldf = self.X_.ldf_.values[..., :len(self.X_.ddims)-1]
        ldf = np.concatenate(
            (ldf, np.prod(self.X_.ldf_.values[..., -extend:], -1,
             keepdims=True)), -1)
        ldf = np.unique(ldf, axis=-2)
        for i in range(self.full_triangle_.shape[-1]-1):
            t_tot = np.sqrt((t1[..., i:i+1])**2 + (ldf[..., i:i+1] *
                            risk_arr[..., -1:])**2)
            risk_arr = np.concatenate((risk_arr, t_tot), -1)
        return risk_arr

    @property
    def mack_std_err_(self):
        obj = copy.copy(self.parameter_risk_)
        obj.values = np.sqrt(self.parameter_risk_.values**2 +
                             self.process_risk_.values**2)
        obj._set_slicers()
        return obj

    @property
    def total_mack_std_err_(self):
        obj = copy.copy(self.X_.latest_diagonal)
        obj.values = np.sqrt(self.total_process_risk_.values**2 +
                             self.total_parameter_risk_.values**2)
        obj.values = obj.values[..., -1:]
        return pd.DataFrame(
            obj.values[..., 0, 0], index=obj.kdims,
            columns=[item + ' Total Mack Std Err' for item in obj.vdims])

    @property
    def summary_(self):
        # This might be better as a dataframe
        obj = copy.copy(self.X_)
        obj.values = np.concatenate(
            (self.X_.latest_diagonal.values,
             self.ibnr_.values,
             self.ultimate_.values,
             self.mack_std_err_.values[..., -1:]), 3)
        obj.ddims = ['Latest', 'IBNR', 'Ultimate', 'Mack Std Err']
        obj.nan_override = True
        obj._set_slicers()
        return obj



import pandas as pd
import numpy as np
import joblib
import json
import os
from chainladder.core.triangle import Triangle
from chainladder.workflow import Pipeline


def load_dataset(key, *args, **kwargs):
    """ Function to load datasets included in the chainladder package.

        Arguments:
        key: str
        The name of the dataset, e.g. RAA, ABC, UKMotor, GenIns, etc.

        Returns:
    	pandas.DataFrame of the loaded dataset.

   """
    path = os.path.dirname(os.path.abspath(__file__))
    origin = 'origin'
    development = 'development'
    columns = ['values']
    index = None
    if key.lower() in ['mcl', 'usaa', 'quarterly', 'auto', 'usauto']:
        columns = ['incurred', 'paid']
    if key.lower() == 'clrd':
        origin = 'AccidentYear'
        development = 'DevelopmentYear'
        index = ['GRNAME', 'LOB']
        columns = ['IncurLoss', 'CumPaidLoss', 'BulkLoss', 'EarnedPremDIR',
                   'EarnedPremCeded', 'EarnedPremNet']
    if key.lower() in ['liab', 'auto']:
        index = ['lob']
    if key.lower() in ['cc_sample', 'ia_sample']:
        columns = ['loss', 'exposure']
    df = pd.read_csv(os.path.join(path, 'data', key.lower() + '.csv'))
    return Triangle(df, origin=origin, development=development, index=index,
                    columns=columns, cumulative=True, *args, **kwargs)


def read_pickle(path):
    return joblib.load(path)


def read_json(json_str):
    json_dict = json.loads(json_str)
    if type(json_dict) is list:
        import chainladder as cl
        return Pipeline(steps=[
            (item['name'],
             cl.__dict__[item['__class__']]().set_params(**item['params']))
            for item in json_dict])
    elif 'kdims' in json_dict.keys():
        tri = Triangle()
        arrays = ['values', 'kdims', 'vdims', 'odims', 'ddims']
        for array in arrays:
            setattr(tri, array, np.array(
                json_dict[array]['array'], dtype=json_dict[array]['dtype']))
        properties = ['key_labels', 'origin_grain', 'development_grain',
                    'nan_override']
        for prop in properties:
            setattr(tri, prop, json_dict[prop])
        tri.valuation_date = pd.to_datetime(
            json_dict['valuation_date'], format='%Y-%m-%d')
        tri._set_slicers()
        tri.valuation = tri._valuation_triangle()
        return tri
    else:
        import chainladder as cl
        return cl.__dict__[
            json_dict['__class__']]().set_params(**json_dict['params'])

    

def parallelogram_olf(values, date, start_date=None, end_date=None,
                      grain='M', vertical_line=False):
    """ Parallelogram approach to on-leveling.  Need to fix return grain
    """
    date = pd.to_datetime(date)
    if not start_date:
        start_date = '{}-01-01'.format(date.min().year-1)
    if not end_date:
        end_date = '{}-12-31'.format(date.max().year+1)
    date_idx = pd.date_range(start_date, end_date)
    y = pd.Series(np.array(values), np.array(date))
    y = y.reindex(date_idx, fill_value=0)
    idx = np.cumprod(y.values+1)
    idx = idx[-1]/idx
    y = pd.Series(idx, y.index)
    if not vertical_line:
        y = y.to_frame().rolling(365).mean()
    y = y.groupby(y.index.to_period(grain)).mean().reset_index()
    y.columns = ['Origin', 'OLF']
    y['Origin'] = y['Origin'].astype(str)
    return y.set_index('Origin')



import pandas as pd
import numpy as np
import copy


class _LocBase:
    ''' Base class for pandas style loc/iloc indexing '''
    def __init__(self, obj):
        self.obj = obj

    def get_idx(self, idx):
        ''' Returns a slice of the original Triangle '''
        obj = copy.deepcopy(self.obj)
        vdims = pd.Series(obj.vdims)
        obj.kdims = np.array(idx.index.unique())
        # Honor order of column labels
        obj.vdims = np.array(vdims[vdims.isin(idx.columns.unique())])
        obj.key_labels = list(idx.index.names)
        obj.iloc, obj.loc = Ilocation(obj), Location(obj)
        idx_slice = np.array(idx).flatten()
        x = tuple([np.unique(np.array(item))
                   for item in list(zip(*idx_slice))])
        obj.values = obj.values[x[0]][:, x[1]]
        obj.values[obj.values == 0] = np.nan
        return obj


class Location(_LocBase):
    ''' class to generate .loc[] functionality '''
    def __getitem__(self, key):
        idx = self.obj._idx_table().loc[key]
        return self.get_idx(self.obj._idx_table_format(idx))


class Ilocation(_LocBase):
    ''' class to generate .iloc[] functionality '''
    def __getitem__(self, key):
        idx = self.obj._idx_table().iloc[key]
        return self.get_idx(self.obj._idx_table_format(idx))


class TriangleSlicer:
    ''' Slicer functionality '''
    def _idx_table_format(self, idx):
        if type(idx) is pd.Series:
            # One row or one column selection is it k or v?
            if len(set(idx.index).intersection(set(self.vdims))) == len(idx):
                # One column selection
                idx = idx.to_frame().T
                idx.index.names = self.key_labels
            else:
                # One row selection
                idx = idx.to_frame()
        elif type(idx) is tuple:
            # Single cell selection
            idx = self._idx_table().iloc[idx[0]:idx[0] + 1,
                                         idx[1]:idx[1] + 1]
        return idx

    def _idx_table(self):
        ''' private method that generates a dataframe of triangle indices.
            The dataframe is meant ot be sliced using pandas and the resultant
            indices are then to be extracted from the Triangle object.
        '''
        df = pd.DataFrame(list(self.kdims), columns=self.key_labels)
        for num, item in enumerate(self.vdims):
            df[item] = list(zip(np.arange(len(df)),
                            (np.ones(len(df))*num).astype(int)))
        df.set_index(self.key_labels, inplace=True)
        return df

    def __getitem__(self, key):
        ''' Function for pandas style column indexing'''
        if type(key) is pd.DataFrame and 'development' in key.columns:
            return self._slice_development(key['development'])
        elif type(key) is np.ndarray:
            # Presumes that if I have a 1D array, I will want to slice origin.
            if len(key) == np.prod(self.shape[-2:]) and self.shape[-1] > 1:
                return self._slice_valuation(key)
            return self._slice_origin(key)
        elif type(key) is pd.Series:
            return self.iloc[list(self.index[key].index)]
        elif key in self.key_labels:
            # Boolean-indexing of a particular key
            return self.index[key]
        else:
            idx = self._idx_table()[key]
            idx = self._idx_table_format(idx)
            obj = _LocBase(self).get_idx(idx)
            if type(key) is not str and key != list(obj.vdims):
                # Honor order of the slice
                obj2 = obj[key[0]]
                for item in key[1:]:
                    obj2[item] = obj[item]
                return obj2
            else:
                return _LocBase(self).get_idx(idx)

    def __setitem__(self, key, value):
        ''' Function for pandas style column indexing setting '''
        idx = self._idx_table()
        idx[key] = 1
        if key in self.vdims:
            i = np.where(self.vdims == key)[0][0]
            self.values[:, i:i+1] = value.values
        else:
            self.vdims = np.array(idx.columns.unique())
            self.values = np.append(self.values, value.values, axis=1)

    def _slice_origin(self, key):
        ''' private method for handling of origin slicing '''
        obj = copy.deepcopy(self)
        obj.odims = obj.odims[key]
        obj.values = obj.values[..., key, :]
        return self._cleanup_slice(obj)

    def _slice_valuation(self, key):
        ''' private method for handling of valuation slicing '''
        obj = copy.deepcopy(self)
        obj.valuation_date = min(
            obj.valuation[key].max().to_timestamp(how='e'), obj.valuation_date)
        key = key.reshape(self.shape[-2:], order='f')
        nan_tri = np.ones(self.shape[-2:])
        nan_tri = key*nan_tri
        nan_tri[nan_tri == 0] = np.nan
        o, d = nan_tri.shape
        o_idx = np.arange(o)[list(np.sum(np.isnan(nan_tri), 1) != d)]
        d_idx = np.arange(d)[list(np.sum(np.isnan(nan_tri), 0) != o)]
        obj.odims = obj.odims[np.sum(np.isnan(nan_tri), 1) != d]
        if len(obj.ddims) > 1:
            obj.ddims = obj.ddims[np.sum(np.isnan(nan_tri), 0) != o]
        obj.values = (obj.values*nan_tri)
        obj.values = np.take(np.take(obj.values, o_idx, -2), d_idx, -1)
        return self._cleanup_slice(obj)

    def _slice_development(self, key):
        ''' private method for handling of development slicing '''
        obj = copy.deepcopy(self)
        obj.ddims = obj.ddims[key]
        obj.values = obj.values[..., key]
        return self._cleanup_slice(obj)

    def _cleanup_slice(self, obj):
        ''' private method with common post-slicing functionality'''
        obj.valuation = obj._valuation_triangle()
        if hasattr(obj, '_nan_triangle_'):
            # Force update on _nan_triangle at next access.
            del obj._nan_triangle_
            obj._nan_triangle_ = obj._nan_triangle()
        return obj

    def _set_slicers(self):
        ''' Call any time the shape of index or column changes '''
        self.iloc, self.loc = Ilocation(self), Location(self)



import numpy as np
import pandas as pd
import copy
from chainladder.methods import MethodBase


class Benktander(MethodBase):
    """ The Benktander (or iterated Bornhuetter-Ferguson) IBNR model

    Parameters
    ----------
    apriori : float, optional (default=1.0)
        Multiplier for the sample_weight used in the Benktander method
        method. If sample_weight is already an apriori measure of ultimate,
        then use 1.0
    n_iters : int, optional (default=1)
        Multiplier for the sample_weight used in the Bornhuetter Ferguson
        method. If sample_weight is already an apriori measure of ultimate,
        then use 1.0

    Attributes
    ----------
    ultimate_ : Triangle
        The ultimate losses per the method
    ibnr_ : Triangle
        The IBNR per the method

    References
    ----------
    .. [2] Benktander, G. (1976) An Approach to Credibility in Calculating IBNR
           for Casualty Excess Reinsurance. In The Actuarial Review, April 1976, p.7
    """
    def __init__(self, apriori=1.0, n_iters=1):
        self.apriori = apriori
        self.n_iters = n_iters

    def fit(self, X, y=None, sample_weight=None):
        """Applies the Benktander technique to triangle **X**

        Parameters
        ----------
        X : Triangle
            Loss data to which the model will be applied.
        y : None
            Ignored
        sample_weight : Triangle
            Required exposure to be used in the calculation.

        Returns
        -------
        self : object
            Returns the instance itself.
        """

        if sample_weight is None:
            raise ValueError('sample_weight is required.')
        super().fit(X, y, sample_weight)
        obj = copy.deepcopy(self)
        self.sample_weight_ = sample_weight
        self.ultimate_ = self._get_ultimate_(X, sample_weight, obj)
        self.full_triangle_ = self._get_full_triangle_()
        return self

    def _get_ultimate_(self, X, sample_weight, obj):
        ult = obj.X_
        origin, development = -2, -1  # Set axes by name
        latest = X.latest_diagonal.values
        apriori = sample_weight.values*self.apriori
        ult.values = \
            obj.cdf_.values[..., :ult.shape[development]]*(ult.values*0+1)
        cdf = ult.latest_diagonal.values
        cdf = (1-1/cdf)[np.newaxis]
        exponents = np.arange(self.n_iters+1)
        exponents = np.reshape(exponents, tuple([len(exponents)]+[1]*4))
        cdf = cdf**exponents
        ult.values = np.sum(cdf[:-1, ...], 0)*latest+cdf[-1, ...]*apriori
        ult.values[~np.isfinite(ult.values)] = np.nan
        ult.ddims = np.array([None])
        ult.valuation = pd.DatetimeIndex([pd.to_datetime('2262-04-11')] *
                                         ult.shape[origin])
        return ult

    def predict(self, X, sample_weight):
        obj = super().predict(X, sample_weight)
        obj.ultimate_ = self._get_ultimate_(X, sample_weight, obj)
        obj.full_triangle_ = obj._get_full_triangle_()
        return obj



""" tails should store all tail methodologies
"""
from chainladder.tails.base import TailBase # noqa (API import)
from chainladder.tails.constant import TailConstant # noqa (API import)
from chainladder.tails.curve import TailCurve # noqa (API import)



import pandas as pd
import numpy as np
import copy


class TriangleDunders:
    ''' Class that implements the dunder (double underscore) methods for the
        Triangle class
    '''
    def _validate_arithmetic(self, other):
        ''' Common functionality BEFORE arithmetic operations '''
        obj = copy.deepcopy(self)
        other = other if type(other) in [int, float] else copy.deepcopy(other)
        ddims = None
        odims = None
        if type(other) not in [int, float, np.float64, np.int64]:
            if len(self.vdims) != len(other.vdims):
                raise ValueError('Triangles must have the same number of ' +
                                 'columns')
            if len(self.kdims) != len(other.kdims):
                raise ValueError('Triangles must have the same number of ' +
                                 'index')
            if len(self.vdims) == 1:
                other.vdims = np.array([None])
            # If broadcasting doesn't work, then try union of origin/developments
            # before failure
            a, b = self.shape[-2:], other.shape[-2:]
            if not (a[0] == 1 or b[0] == 1 or a[0] == b[0]) or \
               not (a[1] == 1 or b[1] == 1 or a[1] == b[1]):
                ddims = pd.concat((pd.Series(self.ddims, index=self.ddims),
                                pd.Series(other.ddims, index=other.ddims)), axis=1)
                odims = pd.concat((pd.Series(self.odims, index=self.odims),
                                pd.Series(other.odims, index=other.odims)), axis=1)
                other_arr = np.zeros(
                    (other.shape[0], other.shape[1], len(odims), len(ddims)))
                other_arr[:] = np.nan
                ol, oh = np.where(~odims[1].isna().values == 1)[0].min(
                ), np.where(~odims[1].isna().values == 1)[0].max()+1
                dl, dh = np.where(~ddims[1].isna().values == 1)[0].min(
                ), np.where(~ddims[1].isna().values == 1)[0].max()+1
                other_arr[:, :, ol:oh, dl:dh] = other.values

                obj_arr = np.zeros(
                    (self.shape[0], self.shape[1], len(odims), len(ddims)))
                obj_arr[:] = np.nan

                ol, oh = np.where(~odims[0].isna().values == 1)[0].min(
                ), np.where(~odims[0].isna().values == 1)[0].max()+1
                dl, dh = np.where(~ddims[0].isna().values == 1)[0].min(
                ), np.where(~ddims[0].isna().values == 1)[0].max()+1
                obj_arr[:, :, ol:oh, dl:dh] = self.values

                odims = np.array(odims.index)
                ddims = np.array(ddims.index)
                obj.ddims =  ddims
                obj.odims =  odims
                obj.values = obj_arr
                other.values = other_arr
                obj._set_slicers()
                obj.valuation = obj._valuation_triangle()
                if hasattr(obj, '_nan_triangle_'):
                    # Force update on _nan_triangle at next access.
                    del obj._nan_triangle_
            other = other.values
        return obj, other

    def _arithmetic_cleanup(self, obj):
        ''' Common functionality AFTER arithmetic operations '''
        obj.values = obj.values * self._expand_dims(obj._nan_triangle())
        obj.values[obj.values == 0] = np.nan
        return obj

    def __add__(self, other):
        obj, other = self._validate_arithmetic(other)
        obj.values = np.nan_to_num(obj.values) + np.nan_to_num(other)
        return self._arithmetic_cleanup(obj)

    def __radd__(self, other):
        return self if other == 0 else self.__add__(other)

    def __sub__(self, other):
        obj, other = self._validate_arithmetic(other)
        obj.values = np.nan_to_num(obj.values) - \
            np.nan_to_num(other)
        return self._arithmetic_cleanup(obj)

    def __rsub__(self, other):
        obj, other = self._validate_arithmetic(other)
        obj.values = np.nan_to_num(other) - \
            np.nan_to_num(obj.values)
        return self._arithmetic_cleanup(obj)

    def __len__(self):
        return self.shape[0]

    def __neg__(self):
        obj = copy.deepcopy(self)
        obj.values = -obj.values
        return obj

    def __pos__(self):
        return self

    def __mul__(self, other):
        obj, other = self._validate_arithmetic(other)
        obj.values = np.nan_to_num(obj.values)*other
        return self._arithmetic_cleanup(obj)

    def __rmul__(self, other):
        return self if other == 1 else self.__mul__(other)

    def __truediv__(self, other):
        obj, other = self._validate_arithmetic(other)
        obj.values = np.nan_to_num(obj.values)/other
        return self._arithmetic_cleanup(obj)

    def __rtruediv__(self, other):
        obj = copy.deepcopy(self)
        obj.values = other / self.values
        obj.values[obj.values == 0] = np.nan
        return obj

    def __eq__(self, other):
        if np.all(np.nan_to_num(self.values) ==
           np.nan_to_num(other.values)):
            return True
        else:
            return False

    def __contains__(self, value):
        if self.__dict__.get(value, None) is None:
            return False
        return True



import chainladder as cl
import numpy as np

def test_non_vertical_line():
    true_olf = (1-.5*(184/365)**2)*.2
    olf_low = cl.parallelogram_olf([.20],['7/1/2017'], grain='Y').loc['2017'].iloc[0]-1
    olf_high = cl.parallelogram_olf([.20],['7/2/2017'], grain='Y').loc['2017'].iloc[0]-1
    assert olf_low < true_olf < olf_high


def test_vertical_line():
    olf = cl.parallelogram_olf([.20], ['7/1/2017'], grain='Y', vertical_line=True)
    assert abs(olf.loc['2017'].iloc[0] - ((1-184/365)*.2+1)) < .00001

def test_triangle_json_io():
    clrd = cl.load_dataset('clrd')
    clrd2 = cl.read_json(clrd.to_json())
    np.testing.assert_equal(clrd.values, clrd2.values)
    np.testing.assert_equal(clrd.kdims, clrd2.kdims)
    np.testing.assert_equal(clrd.vdims, clrd2.vdims)
    np.testing.assert_equal(clrd.odims, clrd2.odims)
    np.testing.assert_equal(clrd.ddims, clrd2.ddims)
    assert np.all(clrd.valuation == clrd2.valuation)

def test_estimator_json_io():
    assert cl.read_json(cl.Development().to_json()).get_params() == \
           cl.Development().get_params()

def test_pipeline_json_io():
    pipe = cl.Pipeline(steps=[('dev', cl.Development()),
                              ('model', cl.BornhuetterFerguson())])
    pipe2 = cl.read_json(pipe.to_json())
    assert {item[0]: item[1].get_params()
            for item in pipe.get_params()['steps']} == \
           {item[0]: item[1].get_params()
            for item in pipe2.get_params()['steps']}



import pandas as pd
import numpy as np
import copy
import inspect



class TriangleGroupBy:
    def __init__(self, old_obj, by):
        obj = copy.deepcopy(old_obj)
        if by != -1:
            indices = obj.index.groupby(by).indices
            new_index = obj.index.groupby(by).count().index
        else:
            indices = {'All': np.arange(len(obj.index))}
            new_index = pd.Index(['All'], name='All')
        groups = [indices[item] for item in sorted(list(indices.keys()))]
        old_k_by_new_k = np.zeros(
            (len(obj.index.index), len(groups)), dtype='bool')
        for num, item in enumerate(groups):
            old_k_by_new_k[:, num][item] = True
        old_k_by_new_k = np.swapaxes(old_k_by_new_k, 0, 1)
        for i in range(3):
            old_k_by_new_k = old_k_by_new_k[..., np.newaxis]
        self.old_k_by_new_k = old_k_by_new_k
        obj.kdims = np.array(list(new_index))
        obj.key_labels = list(new_index.names)
        self.obj = obj

    def quantile(self, q, axis=1, *args, **kwargs):
        """ Return values at the given quantile over requested axis.  If
            Triangle is convertible to DataFrame then pandas quantile
            functionality is used instead.

        Parameters
        ----------
        q: float or array-like, default 0.5 (50% quantile)
            Value between 0 <= q <= 1, the quantile(s) to compute.
        axis:  {0, 1, ‘index’, ‘columns’} (default 1)
            Equals 0 or ‘index’ for row-wise, 1 or ‘columns’ for column-wise.

        Returns
        -------
            Triangle

        """
        x = self.obj.values*self.old_k_by_new_k
        ignore_vector = np.sum(np.isnan(x), axis=1, keepdims=True) == \
            x.shape[1]
        x = np.where(ignore_vector, 0, x)
        self.obj.values = \
            getattr(np, 'nanpercentile')(x, q*100, axis=1, *args, **kwargs)
        self.obj.values[self.obj.values == 0] = np.nan
        return self.obj


class TrianglePandas:
    def to_frame(self, *args, **kwargs):
        """ Converts a triangle to a pandas.DataFrame.  Requires an individual
        index and column selection to appropriately grab the 2D DataFrame.

        Returns
        -------
            pandas.DataFrame representation of the Triangle.
        """
        axes = [num for num, item in enumerate(self.shape) if item > 1]
        if self.shape[:2] == (1, 1):
            return self._repr_format()
        elif len(axes) == 2 or len(axes) == 1:
            tri = np.squeeze(self.values)
            axes_lookup = {0: self.kdims, 1: self.vdims,
                           2: self.origin, 3: self.ddims}
            if len(axes) == 2:
                return pd.DataFrame(tri, index=axes_lookup[axes[0]],
                                    columns=axes_lookup[axes[1]]).fillna(0)
            if len(axes) == 1:
                return pd.Series(tri, index=axes_lookup[axes[0]]).fillna(0)
        else:
            raise ValueError('len(index) and len(columns) must be 1.')

    def plot(self, *args, **kwargs):
        """ Passthrough of pandas functionality """
        return self.to_frame().plot(*args, **kwargs)

    def _get_axis(self, axis):
        ax = {0: 0, 1: 1,2: 2, 3: 3, -1: 3, -2: 2, -3: 1, -4: 0,
             'index': 0,'columns': 1, 'origin': 2, 'development':3}
        return ax.get(axis, 0)

    def dropna(self):
        """  Method that removes orgin/development vectors from edge of a
        triangle that are all missing values. This may come in handy for a
        new line of business that doesn't have origins/developments of an
        existing line in the same triangle.
        """
        obj = self.sum(axis=0).sum(axis=1)
        odim = list(obj.sum(axis=-1).values[0, 0, :, 0]*0+1)
        min_odim = obj.origin[odim.index(1)]
        max_odim = obj.origin[::-1][odim[::-1].index(1)]
        ddim = np.nan_to_num((obj.sum(axis=-2).values*0+1)[0, 0, 0])
        ddim = obj.development.iloc[:, 0][pd.Series(ddim).astype(bool)]
        obj = self[(self.origin >= min_odim) & (self.origin <= max_odim)]
        obj = obj[(self.development >= ddim.min()) &
                  (self.development <= ddim.max())]
        return obj

    @property
    def T(self):
        return self.to_frame().T

    def quantile(self, q, *args, **kwargs):
        if self.shape[:2] == (1, 1):
            return self.to_frame().quantile(q, *args, **kwargs)
        return TriangleGroupBy(self, by=-1).quantile(q, axis=1)

    def groupby(self, by, *args, **kwargs):
        """ Group Triangle by index values.  If the triangle is convertable to a
        DataFrame, then it defaults to pandas groupby functionality.

        Parameters
        ----------
        by: str or list
            The index to group by

        Returns
        -------
            GroupBy object (pandas or Triangle)
        """
        if self.shape[:2] == (1, 1):
            return self.to_frame().groupby(*args, **kwargs)
        return TriangleGroupBy(self, by)

    def append(self, other, index):
        """ Append rows of other to the end of caller, returning a new object.

        Parameters
        ----------
        other : Triangle
            The data to append.
        index:
            The index label(s) to assign the appended data.

        Returns
        -------
            New Triangle with appended data.
        """
        return_obj = copy.deepcopy(self)
        x = pd.DataFrame(list(return_obj.kdims), columns=return_obj.key_labels)
        new_idx = pd.DataFrame([index], columns=return_obj.key_labels)
        x = x.append(new_idx, sort=True)
        x.set_index(return_obj.key_labels, inplace=True)
        return_obj.values = np.append(return_obj.values, other.values, axis=0)
        return_obj.kdims = np.array(x.index.unique())
        return_obj._set_slicers()
        return return_obj

    def rename(self, axis, value):
        """ Alter axes labels.

        Parameters
        ----------
            axis: str or int
                A value of 0 <= axis <= 4 corresponding to axes 'index',
                'columns', 'origin', 'development' respectively.  Both the
                int and str representation can be used.
            value: list or str
                List of new labels to be assigned to the axis. List must be of
                same length of the specified axis.

        Returns
        -------
            Triangle with relabeled axis.
        """
        value = [value] if type(value) is str else value
        if axis == 'index' or axis == 0:
            self.index = value
        if axis == 'columns' or axis == 1:
            self.columns = value
        if axis == 'origin' or axis == 2:
            self.origin = value
        if axis == 'development' or axis == 3:
            self.development = value
        return self

    def astype(self, dtype, inplace=True):
        '''
        Copy of the array, cast to a specified type.

        Parameters
        ----------
            dtype : str or dtype
                Typecode or data-type to which the array is cast.
            copy : bool, optional
                By default, astype always returns a newly allocated array.
        Returns
        -------
            Triangle as new datatype.
        '''
        obj = copy.deepcopy(self) if inplace is True else self
        obj.values = obj.values.astype(dtype)
        return obj


def add_triangle_agg_func(cls, k, v):
    ''' Aggregate Overrides in Triangle '''
    def agg_func(self, axis=None, *args, **kwargs):
            obj = copy.deepcopy(self)
            if axis is None:
                axis = min([num for num, _ in enumerate(obj.shape) if _ != 1])
            else:
                axis = self._get_axis(axis)
            func = getattr(np, v)
            if 'keepdims' in inspect.getfullargspec(func).args:
                obj.values = func(
                    obj.values, axis=axis, keepdims=True, *args, **kwargs)
            else:
                obj.values = func(
                    obj.values, axis=axis, *args, **kwargs)
            if axis == 0 and obj.values.shape[axis] == 1:
                obj.kdims = np.array([None])
                obj.key_labels = [None]
            if axis == 1 and obj.values.shape[axis] == 1:
                obj.vdims = np.array([None])
            if axis == 2 and obj.values.shape[axis] == 1:
                obj.odims = np.array([None])
            if axis == 3 and obj.values.shape[axis] == 1:
                obj.ddims = np.array([None])
            obj._set_slicers()
            obj.values = obj.values * obj._expand_dims(obj._nan_triangle())
            obj.values[obj.values == 0] = np.nan
            if obj.shape == (1, 1, 1, 1):
                return obj.values[0, 0, 0, 0]
            else:
                return obj
    set_method(cls, agg_func, k)


def add_groupby_agg_func(cls, k, v):
    ''' Aggregate Overrides in GroupBy '''
    def agg_func(self, axis=1, *args, **kwargs):
        obj = copy.deepcopy(self.obj)
        x = np.broadcast_to(
            self.obj.values,
            (self.old_k_by_new_k.shape[0], *self.obj.values.shape)) * \
            self.old_k_by_new_k
        ignore_vector = np.sum(np.isnan(x), axis=1, keepdims=True) == \
            x.shape[1]
        x = np.where(ignore_vector, 0, x)
        x[~np.isfinite(x)] = np.nan
        obj.values = \
            getattr(np, v)(x, axis=1, *args, **kwargs)
        obj.values[obj.values == 0] = np.nan
        return obj
    set_method(cls, agg_func, k)


def add_df_passthru(cls, k):
    '''Pass Through of pandas functionality '''
    def df_passthru(self, *args, **kwargs):
        return getattr(pd.DataFrame, k)(self.to_frame(), *args, **kwargs)
    set_method(cls, df_passthru, k)


def set_method(cls, func, k):
    ''' Assigns methods to a class '''
    func.__doc__ = 'Refer to pandas for ``{}`` functionality.'.format(k)
    func.__name__ = k
    setattr(cls, func.__name__, func)


df_passthru = ['to_clipboard', 'to_csv', 'to_excel', 'to_json',
               'to_html', 'to_dict', 'unstack', 'pivot', 'drop_duplicates',
               'describe', 'melt', 'pct_chg']
agg_funcs = ['sum', 'mean', 'median', 'max', 'min', 'prod',
             'var', 'std', 'cumsum']
agg_funcs = {item: 'nan'+item for item in agg_funcs}
more_aggs= ['diff']
agg_funcs = {**agg_funcs, **{item: item for item in more_aggs}}

for item in df_passthru:
    add_df_passthru(TrianglePandas, item)

for k, v in agg_funcs.items():
    add_triangle_agg_func(TrianglePandas, k, v)
    add_groupby_agg_func(TriangleGroupBy, k, v)



import numpy as np
from chainladder.tails import TailBase
from chainladder.development import DevelopmentBase

class TailConstant(TailBase):
    """Allows for the entry of a constant tail factor to LDFs.

    Parameters
    ----------
    tail : float
        The constant to apply to all LDFs within a triangle object.
    decay : float (default=0.50)
        An exponential decay constant that allows for decay over future
        development periods.  A decay rate of 0.5 sets the development portion
        of each successive LDF to 50% of the previous LDF.

    Attributes
    ----------
    ldf_ :
        ldf with tail applied.
    cdf_ :
        cdf with tail applied.
    sigma_ :
        sigma with tail factor applied.
    std_err_ :
        std_err with tail factor applied

    Examples
    --------
    >>> import chainladder as cl
    >>> abc = cl.Development().fit_transform(cl.load_dataset('abc'))
    >>> abc = cl.TailConstant(tail=1.05).fit_transform(abc)
    >>> abc.ldf_
             12-24     24-36     36-48     48-60     60-72     72-84     84-96    96-108   108-120   120-132  132-Ult
    1977  2.308599  1.421098  1.199934  1.113445  1.072736  1.047559  1.034211  1.026047  1.020188  1.016259     1.05
    1978  2.308599  1.421098  1.199934  1.113445  1.072736  1.047559  1.034211  1.026047  1.020188  1.016259     1.05
    1979  2.308599  1.421098  1.199934  1.113445  1.072736  1.047559  1.034211  1.026047  1.020188  1.016259     1.05
    1980  2.308599  1.421098  1.199934  1.113445  1.072736  1.047559  1.034211  1.026047  1.020188  1.016259     1.05
    1981  2.308599  1.421098  1.199934  1.113445  1.072736  1.047559  1.034211  1.026047  1.020188  1.016259     1.05
    1982  2.308599  1.421098  1.199934  1.113445  1.072736  1.047559  1.034211  1.026047  1.020188  1.016259     1.05
    1983  2.308599  1.421098  1.199934  1.113445  1.072736  1.047559  1.034211  1.026047  1.020188  1.016259     1.05
    1984  2.308599  1.421098  1.199934  1.113445  1.072736  1.047559  1.034211  1.026047  1.020188  1.016259     1.05
    1985  2.308599  1.421098  1.199934  1.113445  1.072736  1.047559  1.034211  1.026047  1.020188  1.016259     1.05
    1986  2.308599  1.421098  1.199934  1.113445  1.072736  1.047559  1.034211  1.026047  1.020188  1.016259     1.05
    1987  2.308599  1.421098  1.199934  1.113445  1.072736  1.047559  1.034211  1.026047  1.020188  1.016259     1.05

    Notes
    -----
    The tail constant does not support the entry of variability parameters
    necessary for stochastic approaches, so any usage of TailConstant will be
    inherently deterministic.

    See also
    --------
    TailCurve

    """
    def __init__(self, tail=1.0, decay=0.5):
        self.tail = tail
        self.decay = decay

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the tail will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        super().fit(X, y, sample_weight)
        decay_range = self.ldf_.shape[-1]-X.shape[-1]+1
        ldfs = 1+self._get_initial_ldf()*(self.decay**np.arange(1000))
        ldfs = ldfs[:decay_range]
        ldfs[-1] = self.tail/np.prod(ldfs[:-1])
        ldfs = X._expand_dims(ldfs[np.newaxis])
        self.ldf_.values[..., -decay_range:] = \
            self.ldf_.values[..., -decay_range:]*ldfs
        self.cdf_ = DevelopmentBase._get_cdf(self)
        return self

    def transform(self, X):
        """Transform X.
        The predicted class of an input sample is a vote by the trees in
        the forest, weighted by their probability estimates. That is,
        the predicted class is the one with highest mean probability
        estimate across the trees.

        Parameters
        ----------
        X : Triangle
            Triangle must contain the ``ldf_`` development attribute.

        Returns
        -------
        X_new : Triangle
            New Triangle with tail factor applied to its development
            attributes.
        """
        X.std_err_ = self.std_err_
        X.cdf_ = self.cdf_
        X.ldf_ = self.ldf_
        X.sigma_ = self.sigma_
        return X

    def _get_initial_ldf(self):
        ''' Quadratic series expansion solution '''
        arr = self.decay**np.arange(1000)
        a = np.sum(arr**2)
        b = np.sum(arr)
        c = -np.log(self.tail)
        return (-b+np.sqrt(b**2-4*a*c))/(2*a)



import numpy as np
import copy
from chainladder.methods import MethodBase


class Chainladder(MethodBase):
    """
    The basic determinsitic chainladder method.

    Parameters
    ----------
    None

    Attributes
    ----------
    X_
        returns **X** used to fit the triangle
    ultimate_
        The ultimate losses per the method
    ibnr_
        The IBNR per the method
    full_expectation_
        The ultimates back-filled to each development period in **X** replacing
        the known data
    full_triangle_
        The ultimates back-filled to each development period in **X** retaining
        the known data

    Examples
    --------
    Comparing the ultimates for a company using company LDFs vs industry LDFs

    >>> clrd = cl.load_dataset('clrd')['CumPaidLoss']
    >>> clrd = clrd[clrd['LOB'] == 'wkcomp']
    >>> industry = clrd.sum()
    >>> allstate_industry_cl = cl.Chainladder().fit(industry).predict(clrd).ultimate_.loc['Allstate Ins Co Grp']
    >>> allstate_company_cl = cl.Chainladder().fit(clrd.loc['Allstate Ins Co Grp']).ultimate_
    >>> (allstate_industry_cl - allstate_company_cl).rename(development='Industry to Company LDF Diff')
          Industry to Company LDF Diff
    1988                      0.000000
    1989                   -202.830662
    1990                  -4400.765402
    1991                  -5781.419855
    1992                  -6286.251679
    1993                  -4792.896607
    1994                  -6652.981043
    1995                  -8327.246649
    1996                  -7169.608047
    1997                   -273.269090
    """

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Data to which the model will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        super().fit(X, y, sample_weight)
        self.full_triangle_ = self._get_full_triangle_()
        return self

    def predict(self, X):
        obj = super().predict(X)
        obj.full_triangle_ = obj._get_full_triangle_()
        return obj

    @property
    def ultimate_(self):
        development = -1
        nans = self.X_._nan_triangle()
        obj = copy.copy(self.X_)
        obj.values = np.repeat(self.X_.latest_diagonal.values,
                               self.cdf_.shape[development], development)
        cdf = self.cdf_.values[..., :nans.shape[development]]
        obj_tri = obj.values[..., :nans.shape[development]]
        if np.unique(cdf, axis=2).shape[2] == 1 and \
           len(obj.odims) != cdf.shape[2]:
            cdf = np.repeat(np.unique(cdf, axis=2), len(obj.odims), axis=2)
        obj.values = (cdf*obj_tri)*nans
        obj = obj.latest_diagonal
        obj.ddims = np.array([None])
        obj.valuation = obj._valuation_triangle()
        obj._set_slicers()
        return obj



"""
Loss Development
================
"""
import numpy as np
import pandas as pd
import copy
import warnings
from sklearn.base import BaseEstimator, TransformerMixin
from chainladder import WeightedRegression
from chainladder.core import EstimatorIO


class DevelopmentBase(BaseEstimator, TransformerMixin, EstimatorIO):
    @staticmethod
    def _get_cdf(obj):
        if 'ldf_' not in obj:
            return
        else:
            obj2 = copy.copy(obj.ldf_)
            cdf_ = np.flip(np.cumprod(np.flip(obj2.values, -1), -1), -1)
            obj2.ddims = [item.replace(item[item.find("-")+1:], '9999')
                          for item in obj2.ddims]
            obj2.values = cdf_
            obj2._set_slicers()
            return obj2


class Development(DevelopmentBase):
    """ A Transformer that allows for basic loss development pattern selection.

    Parameters
    ----------
    n_periods : integer, optional (default=-1)
        number of origin periods to be used in the ldf average calculation. For
        all origin periods, set n_periods=-1
    average : string, optional (default='volume')
        type of averaging to use for ldf average calculation.  Options include
        'volume', 'simple', and 'regression'
    sigma_interpolation : string optional (default='log-linear')
        Options include 'log-linear' and 'mack'
    drop : tuple or list of tuples
        Drops specific origin/development combination(s)
    drop_high : bool or list of bool (default=None)
        Drops highest link ratio(s) from LDF calculation
    drop_low : bool or list of bool (default=None)
        Drops lowest link ratio(s) from LDF calculation
    drop_valuation : str or list of str (default=None)
        Drops specific valuation periods. str must be date convertible.


    Attributes
    ----------
    ldf_ : Triangle
        The estimated loss development patterns
    cdf_ : Triangle
        The estimated cumulative development patterns
    sigma_ : Triangle
        Sigma of the ldf regression
    std_err_ : Triangle
        Std_err of the ldf regression
    w_ : Triangle
        The weight used in the ldf regression

    """
    def __init__(self, n_periods=-1, average='volume',
                 sigma_interpolation='log-linear', drop=None,
                 drop_high=None, drop_low=None, drop_valuation=None):
        self.n_periods = n_periods
        self.average = average
        self.sigma_interpolation = sigma_interpolation
        self.drop_high = drop_high
        self.drop_low = drop_low
        self.drop_valuation = drop_valuation
        self.drop = drop

    def _assign_n_periods_weight(self, X):
        if type(self.n_periods) is int:
            return self._assign_n_periods_weight_int(X, self.n_periods)[..., :-1]
        elif type(self.n_periods) is list:
            if len(self.n_periods) != X.values.shape[-1]-1:
                raise ValueError('n_periods list must be of lenth {}.'
                                 .format(X.values.shape[-1]-1))
            else:
                return self._assign_n_periods_weight_list(X)
        else:
            raise ValueError('n_periods must be of type <int> or <list>')

    def _assign_n_periods_weight_list(self, X):
        ''' private method to standardize the n_periods input to a list '''
        dict_map = {item: self._assign_n_periods_weight_int(X, item)
                    for item in set(self.n_periods)}
        conc = [dict_map[item][..., num:num+1, :]
                for num, item in enumerate(self.n_periods)]
        return np.swapaxes(np.concatenate(tuple(conc), -2), -2, -1)

    def _assign_n_periods_weight_int(self, X, n_periods):
        ''' Zeros out weights depending on number of periods desired
            Only works for type(n_periods) == int
        '''
        if n_periods < 1 or n_periods >= X.shape[-2] - 1:
            return X.values*0+1
        else:
            val_offset = {
                'Y': {'Y': 1},
                'Q': {'Y':4, 'Q': 1},
                'M': {'Y':12, 'Q': 3, 'M': 1}}
            val_date_min = \
                X.valuation[X.valuation<=X.valuation_date].drop_duplicates().sort_values()
            val_date_min = \
                val_date_min[-n_periods * \
                val_offset[X.development_grain][X.origin_grain] - 1]
            w = X[X.valuation>=val_date_min]
            return np.nan_to_num((w/w).values)*X._expand_dims(X._nan_triangle())
           

    def _drop_adjustment(self, X, link_ratio):
        weight = X._nan_triangle()[:, :-1]
        if self.drop_high == self.drop_low == \
           self.drop == self.drop_valuation is None:
            return weight
        if self.drop_high is not None:
            weight = weight*self._drop_hilo('high', X, link_ratio)
        if self.drop_low is not None:
            weight = weight*self._drop_hilo('low', X, link_ratio)
        if self.drop is not None:
            weight = weight*self._drop(X)
        if self.drop_valuation is not None:
            weight = weight*self._drop_valuation(X)
        return weight

    def _drop_hilo(self, kind, X, link_ratio):
        link_ratio[link_ratio == 0] = np.nan
        lr_valid_count = np.sum(~np.isnan(link_ratio)[0, 0], axis=0)
        if kind == 'high':
            vals = np.nanmax(link_ratio, -2, keepdims=True)
            drop_hilo = self.drop_high
        else:
            vals = np.nanmin(link_ratio, -2, keepdims=True)
            drop_hilo = self.drop_low
        hilo = 1*(vals != link_ratio)
        if type(drop_hilo) is bool:
            drop_hilo = [drop_hilo]*(len(X.development)-1)
        for num, item in enumerate(self.average_):
            if not drop_hilo[num]:
                hilo[..., num] = hilo[..., num]*0+1
            else:
                if lr_valid_count[num] < 3:
                    hilo[..., num] = hilo[..., num]*0+1
                    warnings.warn('drop_high and drop_low cannot be computed '
                                  'when less than three LDFs are present. '
                                  'Ignoring exclusions in some cases.')
        return hilo

    def _drop_valuation(self, X):
        if type(self.drop_valuation) is not list:
            drop_valuation = [self.drop_valuation]
        else:
            drop_valuation = self.drop_valuation
        arr = 1-np.nan_to_num(X[X.valuation.isin(
            pd.PeriodIndex(drop_valuation,
                           freq=X.origin_grain))].values[0, 0]*0+1)
        ofill = X.shape[-2]-arr.shape[-2]
        dfill = X.shape[-1]-arr.shape[-1]
        if ofill > 0:
            arr = np.concatenate((arr, np.repeat(
                np.ones(arr.shape[-1])[np.newaxis], ofill, 0)), 0)
        if dfill > 0:
            arr = np.concatenate((arr, np.repeat(
                np.ones(arr.shape[-2])[..., np.newaxis], dfill, -1)), -1)
        return arr[:, :-1]

    def _drop(self, X):
        drop = [self.drop] if type(self.drop) is not list else self.drop
        arr = X._nan_triangle()
        for item in drop:
            arr[np.where(X.origin == item[0])[0][0],
                np.where(X.development == item[1])[0][0]] = 0
        return arr[:, :-1]

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the munich adjustment will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        if (type(X.ddims) != np.ndarray):
            raise ValueError('Triangle must be expressed with development lags')
        tri_array = X.values.copy()
        tri_array[tri_array == 0] = np.nan
        if type(self.average) is str:
            average = [self.average] * (tri_array.shape[-1] - 1)
        else:
            average = self.average
        average = np.array(average)
        self.average_ = average
        weight_dict = {'regression': 0, 'volume': 1, 'simple': 2}
        x, y = tri_array[..., :-1], tri_array[..., 1:]
        val = np.array([weight_dict.get(item.lower(), 1)
                        for item in average])
        for i in [2, 1, 0]:
            val = np.repeat(val[np.newaxis], tri_array.shape[i], axis=0)
        val = np.nan_to_num(val * (y * 0 + 1))
        link_ratio = np.divide(y, x, where=np.nan_to_num(x) != 0)
        self.w_ = np.array(self._assign_n_periods_weight(X) *
                           self._drop_adjustment(X, link_ratio),
                           dtype='float16')
        w = self.w_ / (x**(val))
        params = WeightedRegression(axis=2, thru_orig=True).fit(x, y, w)
        if self.n_periods != 1:
            params = params.sigma_fill(self.sigma_interpolation)
        else:
            warnings.warn('Setting n_periods=1 does not allow enough degrees '
                          'of freedom to support calculation of all regression'
                          ' statistics.  Only LDFs have been calculated.')
        params.std_err_ = np.nan_to_num(params.std_err_) + \
            np.nan_to_num(
                (1-np.nan_to_num(params.std_err_*0+1)) *
                params.sigma_ /
                np.swapaxes(np.sqrt(x**(2-val))[..., 0:1, :], -1, -2))
        params = np.concatenate(
            (params.slope_, params.sigma_, params.std_err_), 3)
        params = np.swapaxes(params, 2, 3)
        self.ldf_ = self._param_property(X, params, 0)
        self.cdf_ = self._get_cdf(self)
        self.sigma_ = self._param_property(X, params, 1)
        self.std_err_ = self._param_property(X, params, 2)
        return self

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = copy.copy(X)
        triangles = ['std_err_', 'cdf_', 'ldf_', 'sigma_']
        for item in triangles + ['average_', 'w_', 'sigma_interpolation']:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        return X_new

    def _param_property(self, X, params, idx):
        obj = copy.copy(X)
        obj.values = np.ones(X.shape)[..., :-1]*params[..., idx:idx+1, :]
        obj.ddims = X.link_ratio.ddims
        obj.valuation = obj._valuation_triangle(obj.ddims)
        obj.nan_override = True
        obj._set_slicers()
        return obj



import numpy as np
import pytest
from numpy.testing import assert_allclose
import chainladder as cl
from rpy2.robjects.packages import importr
from rpy2.robjects import r

CL = importr('ChainLadder')


@pytest.fixture
def atol():
    return 1e-5


def mack_r(data, alpha, est_sigma, tail):
    if tail:
        return r('mack<-MackChainLadder({},alpha={}, est.sigma="{}", tail=TRUE)'.format(data, alpha, est_sigma))
    else:
        return r('mack<-MackChainLadder({},alpha={}, est.sigma="{}")'.format(data, alpha, est_sigma))


def mack_p(data, average, est_sigma, tail):
    if tail:
        return cl.MackChainladder().fit(cl.TailCurve(curve='exponential').fit_transform(cl.Development(average=average, sigma_interpolation=est_sigma).fit_transform(cl.load_dataset(data))))
    else:
        return cl.MackChainladder().fit(cl.Development(average=average, sigma_interpolation=est_sigma).fit_transform(cl.load_dataset(data)))


data = ['ABC', 'MW2008']
tail = [True, False]
averages = [('simple', 0), ('volume', 1), ('regression', 2)]
est_sigma = [('log-linear', 'log-linear'), ('mack', 'Mack')]


def test_mack_to_triangle():
    assert cl.MackChainladder().fit(cl.TailConstant().fit_transform(cl.Development().fit_transform(cl.load_dataset('ABC')))).summary_ == \
        cl.MackChainladder().fit(cl.Development().fit_transform(cl.load_dataset('ABC'))).summary_


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages)
@pytest.mark.parametrize('est_sigma', est_sigma)
@pytest.mark.parametrize('tail', tail)
def test_mack_full_std_err(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx('F.se')
    p = mack_p(data, averages[0], est_sigma[0], tail).full_std_err_.values[0, 0, :, :]
    p = p[:, :-1] if not tail else p
    r = np.array(df[0])
    assert_allclose(r, p, atol=atol)


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages)
@pytest.mark.parametrize('est_sigma', est_sigma)
@pytest.mark.parametrize('tail', tail)
def test_mack_process_risk(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx('Mack.ProcessRisk')
    p = mack_p(data, averages[0], est_sigma[0], tail).process_risk_.values[0, 0, :, :]
    p = p[:, :-1] if not tail else p
    r = np.array(df[0])
    assert_allclose(r, p, atol=atol)


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages)
@pytest.mark.parametrize('est_sigma', est_sigma)
@pytest.mark.parametrize('tail', tail)
def test_mack_parameter_risk(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx('Mack.ParameterRisk')
    p = mack_p(data, averages[0], est_sigma[0], tail).parameter_risk_.values[0, 0, :, :]
    p = p[:, :-1] if not tail else p
    r = np.array(df[0])
    assert_allclose(r, p, atol=atol)


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages)
@pytest.mark.parametrize('est_sigma', est_sigma)
@pytest.mark.parametrize('tail', tail)
def test_mack_total_process_risk(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx('Total.ProcessRisk')
    p = mack_p(data, averages[0], est_sigma[0], tail).total_process_risk_.values[0, 0, :, :]
    p = p[:, :-1] if not tail else p
    r = np.expand_dims(np.array(df[0]), 0)
    assert_allclose(r, p, atol=atol)


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages)
@pytest.mark.parametrize('est_sigma', est_sigma)
@pytest.mark.parametrize('tail', tail)
def test_mack_total_parameter_risk(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx('Total.ParameterRisk')
    p = mack_p(data, averages[0], est_sigma[0], tail).total_parameter_risk_.values[0, 0, :, :]
    p = p[:, :-1] if not tail else p
    r = np.expand_dims(np.array(df[0]),0)
    assert_allclose(r, p, atol=atol)


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages)
@pytest.mark.parametrize('est_sigma', est_sigma)
@pytest.mark.parametrize('tail', tail)
def test_mack_mack_std_err_(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx('Mack.S.E')
    p = mack_p(data, averages[0], est_sigma[0], tail).mack_std_err_.values[0, 0, :, :]
    p = p[:, :-1] if not tail else p
    r = np.array(df[0])
    assert_allclose(r, p, atol=atol)



import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal
import chainladder as cl
from rpy2.robjects.packages import importr
from rpy2.robjects import r

CL = importr('ChainLadder')

@pytest.fixture
def atol():
    return 1e-5

def mack_r(data, alpha, est_sigma):
    return r('mack<-MackChainLadder({},alpha={}, est.sigma="{}", tail=TRUE)'.format(data, alpha, est_sigma))


def mack_p(data, average, est_sigma):
    return cl.TailCurve(curve='exponential').fit_transform(cl.Development(average=average, sigma_interpolation=est_sigma).fit_transform(cl.load_dataset(data)))

def mack_p_no_tail(data, average, est_sigma):
    return cl.Development(average=average, sigma_interpolation=est_sigma).fit_transform(cl.load_dataset(data))

data = ['RAA', 'ABC', 'GenIns', 'MW2008', 'MW2014']
# M3IR5 in R fails silently on exponential tail. Python actually computes it.
averages = [('simple', 0), ('volume', 1), ('regression', 2)]
est_sigma = [('mack', 'Mack'), ('log-linear', 'log-linear')]


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages)
@pytest.mark.parametrize('est_sigma', est_sigma)
def test_mack_tail_ldf(data, averages, est_sigma, atol):
    r = np.array(mack_r(data, averages[1], est_sigma[1]).rx('f'))
    p = mack_p(data, averages[0], est_sigma[0]).ldf_.values[0, 0, :, :]
    p = np.concatenate((p[:,:-2], np.prod(p[:, -2:],-1, keepdims=True)), -1)
    p = np.unique(p, axis=-2)
    assert_allclose(r, p, atol=atol)


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages)
@pytest.mark.parametrize('est_sigma', est_sigma)
def test_mack_tail_sigma(data, averages, est_sigma, atol):
    r = np.array(mack_r(data, averages[1], est_sigma[1]).rx('sigma'))
    p = mack_p(data, averages[0], est_sigma[0]).sigma_.values[0, 0, :, :]
    p = np.unique(p, axis=-2)
    assert_allclose(r, p, atol=atol)


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages)
@pytest.mark.parametrize('est_sigma', est_sigma)
def test_mack_tail_std_err(data, averages, est_sigma, atol):
    r = np.array(mack_r(data, averages[1], est_sigma[1]).rx('f.se'))
    p = mack_p(data, averages[0], est_sigma[0]).std_err_.values[0, 0, :, :]
    p = np.unique(p, axis=-2)
    assert_allclose(r, p, atol=atol)


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages[0:1])
@pytest.mark.parametrize('est_sigma', est_sigma[0:1])
def test_tail_doesnt_mutate_std_err(data, averages, est_sigma):
    p = mack_p(data, averages[0], est_sigma[0]).std_err_.values[:, :, :, :-1]
    p_no_tail = mack_p_no_tail(data, averages[0], est_sigma[0]).std_err_.values
    assert_equal(p_no_tail, p)


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages[0:1])
@pytest.mark.parametrize('est_sigma', est_sigma[0:1])
def test_tail_doesnt_mutate_ldf_(data, averages, est_sigma):
    p = mack_p(data, averages[0], est_sigma[0]).ldf_.values[..., :len(cl.load_dataset(data).ddims)-1]
    p_no_tail = mack_p_no_tail(data, averages[0], est_sigma[0]).ldf_.values
    assert_equal(p_no_tail, p)


@pytest.mark.parametrize('data', data)
@pytest.mark.parametrize('averages', averages[0:1])
@pytest.mark.parametrize('est_sigma', est_sigma[0:1])
def test_tail_doesnt_mutate_sigma_(data, averages, est_sigma):
    p = mack_p(data, averages[0], est_sigma[0]).sigma_.values[:, :, :, :-1]
    p_no_tail = mack_p_no_tail(data, averages[0], est_sigma[0]).sigma_.values
    assert_equal(p_no_tail, p)



import chainladder as cl

def test_struhuss():
    X = cl.load_dataset('cc_sample')['loss']
    X = cl.TailConstant(tail=1/0.85).fit_transform(cl.Development().fit_transform(X))
    sample_weight = cl.load_dataset('cc_sample')['exposure'].latest_diagonal
    ibnr = int(cl.CapeCod(trend=0.07, decay=0.75).fit(X, sample_weight=sample_weight).ibnr_.sum())
    assert ibnr == 17052



import copy
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from chainladder.development import DevelopmentBase, Development
from chainladder.core import EstimatorIO


class TailBase(BaseEstimator, TransformerMixin, EstimatorIO):
    ''' Base class for all tail methods.  Tail objects are equivalent
        to development objects with an additional set of tail statistics'''
    def fit(self, X, y=None, sample_weight=None):
        obj = copy.copy(X)
        if 'ldf_'not in obj:
            obj = Development().fit_transform(obj)
        self._ave_period = {'Y': (1, 12),
                            'Q': (4, 3),
                            'M': (12, 1)}[obj.development_grain]
        ddims = np.concatenate(
            (obj.ddims, [(item+1)*self._ave_period[1] + obj.ddims[-1]
                         for item in range(self._ave_period[0])], [9999]), 0)
        self.ldf_ = copy.copy(obj.ldf_)
        tail = np.ones(self.ldf_.shape)[..., -1:]
        tail = np.repeat(tail, self._ave_period[0]+1, -1)
        self.ldf_.values = np.concatenate((self.ldf_.values, tail), -1)
        self.ldf_.ddims = np.array(['{}-{}'.format(ddims[i], ddims[i+1])
                                    for i in range(len(ddims)-1)])
        self.ldf_.valuation = self.ldf_._valuation_triangle()
        self.sigma_ = copy.copy(getattr(obj, 'sigma_', obj.cdf_*0))
        self.std_err_ = copy.copy(getattr(obj, 'std_err_', obj.cdf_*0))
        zeros = tail[..., -1:]*0
        self.sigma_.values = np.concatenate(
            (self.sigma_.values, zeros), -1)
        self.std_err_.values = np.concatenate(
            (self.std_err_.values, zeros), -1)
        self.sigma_.ddims = self.std_err_.ddims = \
            np.append(obj.ldf_.ddims, ['{}-9999'.format(int(obj.ddims[-1]))])
        val_array = self.sigma_._valuation_triangle(self.sigma_.ddims)
        self.sigma_.valuation = self.std_err_.valuation = val_array
        self.cdf_ = DevelopmentBase._get_cdf(self)
        self.cdf_._set_slicers()
        self.ldf_._set_slicers()
        self.sigma_._set_slicers()
        self.std_err_._set_slicers()
        return self

    def transform(self, X):
        X_new = copy.copy(X)
        X_new.std_err_.values = np.concatenate(
            (X_new.std_err_.values,
             self.std_err_.values[..., -1:]), -1)
        X_new.cdf_.values = np.concatenate(
            (X_new.cdf_.values,
             self.cdf_.values[..., -self._ave_period[0]-1:]*0+1), -1)
        X_new.cdf_.values = X_new.cdf_.values * \
            self.cdf_.values[..., -self._ave_period[0]-1:-self._ave_period[0]]
        X_new.cdf_.values[..., -1] = self.cdf_.values[..., -1]
        X_new.ldf_.values = np.concatenate(
            (X_new.ldf_.values,
             self.ldf_.values[..., -self._ave_period[0]-1:]), -1)
        X_new.sigma_.values = np.concatenate(
            (X_new.sigma_.values, self.sigma_.values[..., -1:]), -1)
        X_new.cdf_.ddims = X_new.ldf_.ddims = self.ldf_.ddims
        X_new.sigma_.ddims = X_new.std_err_.ddims = self.sigma_.ddims
        X_new.cdf_.valuation = X_new.ldf_.valuation = self.ldf_.valuation
        X_new.sigma_.valuation = \
            X_new.std_err_.valuation = self.sigma_.valuation
        X_new.sigma_._set_slicers()
        X_new.ldf_._set_slicers()
        X_new.cdf_._set_slicers()
        X_new.std_err_._set_slicers()
        return X_new

    def fit_transform(self, X, y=None, sample_weight=None):
        """ Equivalent to fit(X).transform(X)

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs based on the model.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        self.fit(X)
        if 'std_err_' not in X:
            X = Development().fit_transform(X)
        return self.transform(X)

    def _get_tail_prediction(self, tail_ldf):
        ave = 1 + tail_ldf[..., :self._ave_period[0]]
        all = np.expand_dims(
            np.product(1 + tail_ldf[..., self._ave_period[0]:], -1), -1)
        tail = np.concatenate((ave, all), -1)
        return tail



import chainladder as cl
import pandas as pd
import numpy as np
from numpy.testing import assert_equal
import copy

tri = cl.load_dataset('clrd')
qtr = cl.load_dataset('quarterly')

# Test Triangle slicing
def test_slice_by_boolean():
    assert tri[tri['LOB'] == 'ppauto'].loc['Wolverine Mut Ins Co']['CumPaidLoss'] == \
                            tri.loc['Wolverine Mut Ins Co'].loc['ppauto']['CumPaidLoss']


def test_slice_by_loc():
    assert tri.loc['Aegis Grp'].loc['comauto'].index.iloc[0, 0] == 'comauto'


def test_slice_origin():
    assert cl.load_dataset('raa')[cl.load_dataset('raa').origin>'1985'].shape == \
        (1, 1, 5, 10)


def test_slice_development():
    assert cl.load_dataset('raa')[cl.load_dataset('raa').development<72].shape == \
        (1, 1, 10, 5)


def test_slice_by_loc_iloc():
    assert tri.groupby('LOB').sum().loc['comauto'].index.iloc[0, 0] == 'comauto'


def test_repr():
    np.testing.assert_equal(pd.read_html(cl.load_dataset('raa')._repr_html_())[0].set_index('Origin').values,
                            cl.load_dataset('raa').to_frame().values)


def test_arithmetic_union():
    raa = cl.load_dataset('raa')
    assert raa.shape == (raa-raa[raa.valuation<'1987']).shape


def test_to_frame_unusual():
    a = cl.load_dataset('clrd').groupby(['LOB']).sum().latest_diagonal['CumPaidLoss'].to_frame().values
    b = cl.load_dataset('clrd').latest_diagonal['CumPaidLoss'].groupby(['LOB']).sum().to_frame().values
    np.testing.assert_equal(a, b)


def test_link_ratio():
    np.testing.assert_allclose(cl.load_dataset('RAA').link_ratio.values*cl.load_dataset('RAA').values[:,:,:-1,:-1],
                               cl.load_dataset('RAA').values[:,:,:-1,1:], atol=1e-5)


def test_incr_to_cum():
    np.testing.assert_equal(tri.cum_to_incr().incr_to_cum().values, tri.values)


def test_create_new_value():
    tri2 = copy.deepcopy(tri)
    tri2['lr'] = (tri2['CumPaidLoss']/tri2['EarnedPremDIR'])
    assert (tri.shape[0], tri.shape[1]+1, tri.shape[2], tri.shape[3]) == tri2.shape


def test_multilevel_index_groupby_sum1():
    assert tri.groupby('LOB').sum().sum() == tri.sum()


def test_multilevel_index_groupby_sum2():
    a = tri.groupby('GRNAME').sum().sum()
    b = tri.groupby('LOB').sum().sum()
    assert a == b


def test_boolean_groupby_eq_groupby_loc():
    np.testing.assert_equal(tri[tri['LOB']=='ppauto'].sum().values,
                        tri.groupby('LOB').sum().loc['ppauto'].values)


def test_latest_diagonal_two_routes():
    assert tri.latest_diagonal.sum()['BulkLoss'] == tri.sum().latest_diagonal['BulkLoss']


def test_sum_of_diff_eq_diff_of_sum():
    assert (tri['BulkLoss']-tri['CumPaidLoss']).latest_diagonal == \
           (tri.latest_diagonal['BulkLoss'] - tri.latest_diagonal['CumPaidLoss'])


def test_append():
    assert cl.load_dataset('raa').append(cl.load_dataset('raa'), index='raa2').sum() == 2*cl.load_dataset('raa')


def test_assign_existing_col():
    tri = cl.load_dataset('quarterly')
    before = tri.shape
    tri['paid'] = 1/tri['paid']
    assert tri.shape == before


def test_arithmetic_across_keys():
    x = cl.load_dataset('auto')
    np.testing.assert_equal((x.sum()-x.iloc[0]).values, x.iloc[1].values)

def test_grain():
    actual = qtr.iloc[0,0].grain('OYDY').values[0,0,:,:]
    expected = np.array([[  44.,  621.,  950., 1020., 1070., 1069., 1089., 1094., 1097.,
        1099., 1100., 1100.],
       [  42.,  541., 1052., 1169., 1238., 1249., 1266., 1269., 1296.,
        1300., 1300.,   np.nan],
       [  17.,  530.,  966., 1064., 1100., 1128., 1155., 1196., 1201.,
        1200.,   np.nan,   np.nan],
       [  10.,  393.,  935., 1062., 1126., 1209., 1243., 1286., 1298.,
          np.nan,   np.nan,   np.nan],
       [  13.,  481., 1021., 1267., 1400., 1476., 1550., 1583.,   np.nan,
          np.nan,   np.nan,   np.nan],
       [   2.,  380.,  788.,  953., 1001., 1030., 1066.,   np.nan,   np.nan,
          np.nan,   np.nan,   np.nan],
       [   4.,  777., 1063., 1307., 1362., 1411.,   np.nan,   np.nan,   np.nan,
          np.nan,   np.nan,   np.nan],
       [   2.,  472., 1617., 1818., 1820.,   np.nan,   np.nan,   np.nan,   np.nan,
          np.nan,   np.nan,   np.nan],
       [   3.,  597., 1092., 1221.,   np.nan,   np.nan,   np.nan,   np.nan,   np.nan,
          np.nan,   np.nan,   np.nan],
       [   4.,  583., 1212.,   np.nan,   np.nan,   np.nan,   np.nan,   np.nan,   np.nan,
          np.nan,   np.nan,   np.nan],
       [  21.,  422.,   np.nan,   np.nan,   np.nan,   np.nan,   np.nan,   np.nan,   np.nan,
          np.nan,   np.nan,   np.nan],
       [  13.,   np.nan,   np.nan,   np.nan,   np.nan,   np.nan,   np.nan,   np.nan,   np.nan,
          np.nan,   np.nan,   np.nan]])
    np.testing.assert_equal(actual, expected)

def test_off_cycle_val_date():
    assert cl.load_dataset('quarterly').valuation_date.strftime('%Y-%m-%d') == '2006-03-31'

def test_printer():
    print(cl.load_dataset('abc'))


def test_value_order():
    a = tri[['CumPaidLoss','BulkLoss']]
    b = tri[['BulkLoss', 'CumPaidLoss']]
    assert_equal(a.values[:,-1], b.values[:, 0])


def test_trend():
    assert abs((cl.load_dataset('abc').trend(0.05).trend((1/1.05)-1) -
                cl.load_dataset('abc')).sum().sum()) < 1e-5


def test_arithmetic_1():
    x = cl.load_dataset('mortgage')
    np.testing.assert_equal(-(((x/x)+0)*x), -(+x))


def test_arithmetic_2():
    x = cl.load_dataset('mortgage')
    np.testing.assert_equal(1-(x/x), 0*x*0)


def test_rtruediv():
    assert np.nansum(abs(((1/cl.load_dataset('raa'))*cl.load_dataset('raa')).values[0,0] - cl.load_dataset('raa')._nan_triangle()))< .00001


def test_shift():
    x = cl.load_dataset('quarterly').iloc[0,0]
    np.testing.assert_equal(x[x.valuation<=x.valuation_date].values, x.values)

def test_quantile_vs_median():
    clrd = cl.load_dataset('clrd')
    np.testing.assert_equal(clrd.quantile(.5)['CumPaidLoss'].values,
                            clrd.median()['CumPaidLoss'].values)


def test_grain_returns_valid_tri():
    tri = cl.load_dataset('quarterly')
    assert tri.grain('OYDY').latest_diagonal == tri.latest_diagonal


def test_base_minimum_exposure_triangle():
    raa = (cl.load_dataset('raa').latest_diagonal*0+50000).to_frame().reset_index()
    raa['index'] = raa['index'].astype(str)
    cl.Triangle(raa, origin='index',
                columns=list(cl.load_dataset('raa').columns))


def test_origin_and_value_setters():
    raa = cl.load_dataset('raa')
    raa2 = cl.load_dataset('raa')
    raa.columns = list(raa.columns)
    raa.origin = list(raa.origin)
    assert np.all((np.all(raa2.origin == raa.origin),
                   np.all(raa2.development == raa.development),
                   np.all(raa2.odims == raa.odims),
                   np.all(raa2.vdims == raa.vdims)))

def test_grain_increm_arg():
    tri = cl.load_dataset('quarterly')['incurred']
    tri_i = tri.cum_to_incr()
    np.testing.assert_equal(tri_i.grain('OYDY').incr_to_cum(),
                            tri.grain('OYDY'))


def test_valdev1():
    a = cl.load_dataset('quarterly').dev_to_val().val_to_dev().values
    b = cl.load_dataset('quarterly').values
    np.testing.assert_equal(a,b)


def test_valdev2():
    a = cl.load_dataset('quarterly').dev_to_val().grain('OYDY').val_to_dev().values
    b = cl.load_dataset('quarterly').grain('OYDY').values
    np.testing.assert_equal(a,b)


def test_valdev3():
    a = cl.load_dataset('quarterly').grain('OYDY').dev_to_val().val_to_dev().values
    b = cl.load_dataset('quarterly').grain('OYDY').values
    np.testing.assert_equal(a,b)


#def test_valdev4():
#    # Does not work with pandas 0.23, consider requiring only pandas>=0.24
#    raa = cl.load_dataset('raa')
#    np.testing.assert_equal(raa.dev_to_val()[raa.dev_to_val().development>='1989'].values,
#        raa[raa.valuation>='1989'].dev_to_val().values)


def test_valdev5():
    raa = cl.load_dataset('raa')
    np.testing.assert_equal(raa[raa.valuation>='1989'].latest_diagonal.values,
                            raa.latest_diagonal.values)

def test_valdev6():
    np.testing.assert_equal(cl.load_dataset('raa').grain('OYDY').latest_diagonal.values,
                            cl.load_dataset('raa').latest_diagonal.grain('OYDY').values)

def test_valdev7():
    tri = cl.load_dataset('quarterly')
    x = cl.Chainladder().fit(tri).full_expectation_
    np.testing.assert_equal(x.dev_to_val().val_to_dev().values, x.values)

def test_reassignment():
    raa = cl.load_dataset('clrd')
    raa['values'] = raa['CumPaidLoss']
    raa['values'] = raa['values'] + raa['CumPaidLoss']

def test_dropna():
    clrd = cl.load_dataset('clrd')
    assert clrd.shape == clrd.dropna().shape
    assert clrd[clrd['LOB']=='wkcomp'].iloc[-5]['CumPaidLoss'].dropna().shape == (1,1,2,2)

def test_commutative():
    tri = cl.load_dataset('quarterly')
    full = cl.Chainladder().fit(tri).full_expectation_
    assert tri.grain('OYDY').val_to_dev() == tri.val_to_dev().grain('OYDY')
    assert tri.cum_to_incr().grain('OYDY').val_to_dev() == tri.val_to_dev().cum_to_incr().grain('OYDY')
    assert tri.grain('OYDY').cum_to_incr().val_to_dev().incr_to_cum() == tri.val_to_dev().grain('OYDY')
    assert full.grain('OYDY').val_to_dev() == full.val_to_dev().grain('OYDY')
    assert full.cum_to_incr().grain('OYDY').val_to_dev() == full.val_to_dev().cum_to_incr().grain('OYDY')
    assert np.allclose(np.nan_to_num(full.grain('OYDY').cum_to_incr().val_to_dev().incr_to_cum().values),
            np.nan_to_num(full.val_to_dev().grain('OYDY').values), atol=1e-5)



from chainladder.methods import Benktander


class BornhuetterFerguson(Benktander):
    """The deterministic Bornhuetter Ferguson IBNR model

    Parameters
    ----------
    apriori : float, optional (default=1.0)
        Multiplier for the sample_weight used in the Bornhuetter Ferguson
        method. If sample_weight is already an apriori measure of ultimate,
        then use 1.0

    Attributes
    ----------
    ultimate_ : Triangle
        The ultimate losses per the method
    ibnr_ : Triangle
        The IBNR per the method

    Examples
    --------
    Smoothing chainladder ultimates by using them as apriori figures in the
    Bornhuetter Ferguson method.

    >>> raa = cl.load_dataset('RAA')
    >>> cl_ult = cl.Chainladder().fit(raa).ultimate_ # Chainladder Ultimate
    >>> apriori = cl_ult*0+(cl_ult.sum()/10)[0] # Mean Chainladder Ultimate
    >>> cl.BornhuetterFerguson(apriori=1).fit(raa, sample_weight=apriori).ultimate_
              Ultimate
    1981  18834.000000
    1982  16898.632172
    1983  24012.333266
    1984  28281.843524
    1985  28203.700714
    1986  19840.005163
    1987  18840.362337
    1988  22789.948877
    1989  19541.155136
    1990  20986.022826

    References
    ----------
    .. [1] Bornhuetter, R. and Ferguson, R. (1972) The Actuary and IBNR. In
           Proceedings of the Casualty Actuarial Society, Vol. LIX, 181 - 195
    """
    def __init__(self, apriori=1.0):
        self.apriori = apriori

    def fit(self, X, y=None, sample_weight=None):
        """Applies the Bornhuetter-Ferguson technique to triangle **X**

        Parameters
        ----------
        X : Triangle
            Loss data to which the model will be applied.
        y : None
            Ignored
        sample_weight : Triangle
            Required exposure to be used in the calculation.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self.n_iters = 1
        super().fit(X, y, sample_weight)
        return self



import numpy as np
import copy
from sklearn.base import BaseEstimator
from chainladder.tails import TailConstant
from chainladder.development import Development
from chainladder.core import EstimatorIO


class MethodBase(BaseEstimator, EstimatorIO):
    def __init__(self):
        pass

    def validate_X(self, X):
        obj = copy.copy(X)
        if 'ldf_' not in obj:
            obj = Development().fit_transform(obj)
        if len(obj.ddims) - len(obj.ldf_.ddims) == 1:
            obj = TailConstant().fit_transform(obj)
        for item in ['cdf_', 'ldf_', 'average_']:
            setattr(self, item, getattr(obj, item, None))
        return obj

    def fit(self, X, y=None, sample_weight=None):
        """Applies the chainladder technique to triangle **X**

        Parameters
        ----------
        X : Triangle
            The data used to compute the mean and standard deviation
            used for later scaling along the features axis.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self.X_ = self.validate_X(X)
        return self

    def predict(self, X, sample_weight=None):
        """Predicts the chainladder ultimate on a new triangle **X**

        Parameters
        ----------
        X : Triangle
            The data used to compute the mean and standard deviation
            used for later scaling along the features axis.
        sample_weight : Triangle
            For exposure-based methods, the exposure to be used for predictions

        Returns
        -------
        X_new: Triangle

        """
        obj = copy.copy(self)
        obj.X_ = copy.copy(X)
        obj.sample_weight = sample_weight
        if np.unique(self.cdf_.values, axis=-2).shape[-2] == 1:
            obj.cdf_.values = np.repeat(
                np.unique(self.cdf_.values, axis=-2),
                len(X.odims), -2)
            obj.ldf_.values = np.repeat(
                np.unique(self.ldf_.values, axis=-2),
                len(X.odims), -2)
            obj.cdf_.odims = obj.ldf_.odims = obj.X_.odims
            obj.cdf_.valuation = obj.ldf_.valuation = \
                Development().fit(X).cdf_.valuation
        obj.cdf_._set_slicers()
        obj.ldf_._set_slicers()
        return obj

    @property
    def full_expectation_(self):
        obj = copy.copy(self.X_)
        obj.values = (self.ultimate_.values /
                      np.unique(self.cdf_.values, axis=-2))
        obj.values = np.concatenate((obj.values,
                                    self.ultimate_.values), -1)
        ddims = [int(item[item.find('-')+1:]) for item in self.ldf_.ddims]
        obj.ddims = np.array([obj.ddims[0]]+ddims)
        obj.valuation = obj._valuation_triangle(obj.ddims)
        obj.valuation_date = max(obj.valuation).to_timestamp()
        obj.nan_override = True
        obj.values[obj.values == 0] = np.nan
        obj._set_slicers()
        return obj

    def ultimate_(self):
        raise NotImplementedError

    @property
    def ibnr_(self):
        obj = copy.copy(self.ultimate_)
        obj.values = self.ultimate_.values-self.X_.latest_diagonal.values
        obj.ddims = [None]
        obj._set_slicers()
        return obj

    def _get_full_triangle_(self):
        obj = copy.copy(self.X_)
        w = 1-np.nan_to_num(obj._nan_triangle())
        extend = len(self.ldf_.ddims) - len(self.X_.ddims)
        ones = np.ones((w.shape[-2], extend))
        w = np.concatenate((w, ones), -1)
        obj.nan_override = True
        e_tri = \
            np.repeat(self.ultimate_.values, self.cdf_.values.shape[3], 3) / \
            np.unique(self.cdf_.values, axis=-2)
        e_tri = e_tri * w
        zeros = obj._expand_dims(ones - ones)
        properties = self.full_expectation_
        obj.valuation = properties.valuation
        obj.valuation_date = properties.valuation_date
        obj.ddims = properties.ddims
        obj.values = \
            np.concatenate((np.nan_to_num(obj.values), zeros), -1) + e_tri
        obj.values = np.concatenate((obj.values,
                                     self.ultimate_.values), 3)
        obj.values[obj.values==0] = np.nan
        obj._set_slicers()
        return obj



""" utils should store all utility functions and classes, i.e. things that
    are used by various modules in the package.
"""
from chainladder.utils.weighted_regression import WeightedRegression # noqa (API import)
from chainladder.utils.exhibits import (
    DataFrame, Row, Column, Tabs) # noqa (API import)
from chainladder.utils.utility_functions import ( # noqa (API import)
    load_dataset, parallelogram_olf, read_pickle, read_json)



from chainladder.methods.base import MethodBase # noqa (API import)
from chainladder.methods.chainladder import Chainladder # noqa (API import)
from chainladder.methods.mack import MackChainladder # noqa (API import)
from chainladder.methods.benktander import Benktander # noqa (API import)
from chainladder.methods.bornferg import BornhuetterFerguson # noqa (API import)
from chainladder.methods.capecod import CapeCod # noqa (API import)
from chainladder.methods.clark_capecod import ClarkCapeCod # noqa (API import)



from chainladder.utils import * # noqa (API Import)
from chainladder.core import * # noqa (API Import)
from chainladder.development import * # noqa (API Import)
from chainladder.tails import * # noqa (API Import)
from chainladder.methods import * # noqa (API Import)
from chainladder.workflow import * # noqa (API Import)

__version__ = '0.4.4'



"""
Bootstrap Chainladder
=====================
* Generalize to a full set of 4d triangles and not jsut a 1x1xmxn
* Need to include DoF adjustment when hat matrix fails
    - perhaps options for the user to select
* Need to handle first diagonal being cumulative
* Need to develop heteroskedasticity adjustment
* Need to conceptualize how this operates with tail classes

"""

from sklearn.utils import check_random_state

from chainladder.methods.chainladder import Chainladder
from chainladder.development.base import DevelopmentBase, Development
import numpy as np
import pandas as pd
import copy


class BootstrapODPSample(DevelopmentBase):
    """
    Class to generate bootstrap samples of triangles.  Currently this Only
    supports 'single' triangles (single index and single column).

    Parameters
    ----------
    n_sims : int (default=1000)
        Number of simulations to generate
    n_periods : integer, optional (default=-1)
        number of origin periods to be used in the ldf average calculation. For
        all origin periods, set n_periods=-1
    hat_adj : bool (default=TRUE)
        Adjust standardized Pearson residuals with the hat matrix adjustment
        factor.
    drop : tuple or list of tuples
        Drops specific origin/development combination(s) from residual sample
    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by np.random.

    Attributes
    ----------
    resampled_triangles_ : Triangle
        A set of triangles represented by each simulation
    scale_ :
        The scale parameter to be used in generating process risk
    """
    def __init__(self, n_sims=1000, n_periods=-1,
                 hat_adj=True, drop=None, random_state=None):
        self.n_sims = n_sims
        self.n_periods = n_periods
        self.hat_adj = hat_adj
        self.drop = drop
        self.random_state = random_state

    def fit(self, X, y=None, sample_weight=None):
        if (type(X.ddims) != np.ndarray):
            raise ValueError('Triangle must be expressed with development lags')
        obj = copy.copy(X)
        self.w_ = X._nan_triangle() if not self.drop else self._drop(X)
        lag = {'M': 1, 'Q': 3, 'Y': 12}[X.development_grain]
        if type(self.drop) is not list and self.drop is not None:
            self.drop = [self.drop]
            drop = [(item[0], item[1]-lag) for item in self.drop]
        else:
            drop = self.drop
        obj = Development(n_periods=self.n_periods, drop=drop).fit_transform(obj)
        obj = Chainladder().fit(obj)
        # Works for only a single triangle - can we generalize this
        exp_incr_triangle = obj.full_expectation_.cum_to_incr() \
                               .values[0, 0, :, :X.shape[-1]]
        exp_incr_triangle = np.nan_to_num(exp_incr_triangle) * \
            obj.X_._nan_triangle()
        self.design_matrix_ = self._get_design_matrix(X)
        self.hat_ = self._get_hat(X, exp_incr_triangle) if self.hat_adj else None 
        self.resampled_triangles_, self.scale_ = \
            self._get_simulation(X, exp_incr_triangle)
        n_obs = np.nansum(self.w_)
        n_origin_params = X.shape[2]
        n_dev_params = X.shape[3] - 1
        deg_free = n_obs - n_origin_params - n_dev_params
        deg_free_adj_fctr = np.sqrt(n_obs/deg_free)
        return self

    def _get_simulation(self, X, exp_incr_triangle):
        k_value = 1  # for ODP Poisson
        if X.shape[:2] != (1, 1):
            raise ValueError('Only single index/column triangles are ',
                             'supported')
        unscaled_residuals = \
            ((X.cum_to_incr().values - exp_incr_triangle) /
             np.sqrt(np.abs(exp_incr_triangle**k_value)))[0, 0, ...]
        standardized_residuals = self.hat_ * unscaled_residuals
        pearson_chi_sq = sum(sum(np.nan_to_num(unscaled_residuals)**2))
        n_params = self.design_matrix_.shape[1]
        degree_freedom = np.nansum(X._nan_triangle()) - n_params
        # Shapland has a hetero adjustment to degree_freedom here
        # He also adjusts the residuals for the etero adjustment
        scale_phi = pearson_chi_sq/degree_freedom
        k, v, o, d = X.shape
        resids = np.reshape(standardized_residuals, (k, v, o*d))

        adj_resid_dist = resids[np.isfinite(resids)]  # Missing k,v dimensions
        # Suggestions from Using the ODP Bootstrap Model: A Practitioners Guide
        adj_resid_dist = adj_resid_dist[adj_resid_dist != 0]
        adj_resid_dist = adj_resid_dist - np.mean(adj_resid_dist)

        random_state = check_random_state(self.random_state)
        resampled_residual = [random_state.choice(adj_resid_dist,
                              size=exp_incr_triangle.shape,
                              replace=True)*(exp_incr_triangle*0+1)
                              for item in range(self.n_sims)]
        resampled_residual = np.array(resampled_residual) \
                               .reshape(self.n_sims, exp_incr_triangle.shape[0],
                                        exp_incr_triangle.shape[1])
        resampled_residual = resampled_residual
        b = np.repeat(np.expand_dims(exp_incr_triangle, 0), self.n_sims, 0)
        resampled_triangles = (resampled_residual*np.sqrt(abs(b))+b).cumsum(2)
        resampled_triangles = np.swapaxes(
            np.expand_dims(resampled_triangles, 0), 0, 1)
        obj = copy.copy(X)
        obj.kdims = np.arange(self.n_sims)
        obj.values = resampled_triangles
        obj._set_slicers()
        return obj, scale_phi

        # Shapland cites Verral and England 2002 in using gamma as a proxy for
        # poisson because of computational efficiency even though poisson is
        # the more theoretically correct choice.  Does this belong in the fit?

        # Process variance adjustment
        #lower_diagonal = (obj._nan_triangle()*0)
        #lower_diagonal[np.isnan(lower_diagonal)]=1
        #obj.values = Chainladder().fit(Development().fit_transform(obj)).full_expectation_.values[...,:-1]*lower_diagonal
        #obj.nan_override = True
        #if self.process_dist == 'gamma':
        #    process_triangle = np.nan_to_num(np.array([random_state.gamma(shape=abs(item/scale_phi),scale=scale_phi)*np.sign(np.nan_to_num(item)) for item in sim_exp_incr_triangle]))
        #if self.process_dist == 'od poisson':
        #    process_triangle = np.nan_to_num(np.array([random_state.poisson(lam=abs(item))*np.sign(np.nan_to_num(item))for item in sim_exp_incr_triangle]))
        #IBNR = process_triangle.cumsum(axis=2)[:,:,-1]

    def _get_design_matrix(self, X):
        """ The design matrix used in hat matrix adjustment (Shapland eq3.12)
        """
        w = X._nan_triangle()
        arr = np.diag(w[:, 0])
        intra_beta = np.zeros((w.shape[0], w.shape[1]-1))
        arr = np.concatenate((arr, intra_beta), axis=1)
        for i in range(w.shape[1]-1):
            len_alpha = len(w[:, i+1][~np.isnan(w[:, i+1])])
            intra_alpha = np.diag(w[:, i+1])[:len_alpha, :]
            intra_beta[:, i] = 1
            intra_beta = intra_beta[:len_alpha, :]
            intra_arr = np.concatenate((intra_alpha, intra_beta), axis=1)
            arr = np.concatenate((arr, intra_arr), axis=0)
        return arr

    def _get_hat(self, X, exp_incr_triangle):
        """ The hat matrix adjustment (Shapland eq3.23)"""
        weight_matrix = np.diag(pd.DataFrame(exp_incr_triangle).unstack().dropna().values)
        design_matrix = self.design_matrix_
        hat = np.matmul(np.matmul(np.matmul(design_matrix,np.linalg.inv(np.matmul(design_matrix.T, np.matmul(weight_matrix, design_matrix)))), design_matrix.T), weight_matrix)
        hat = np.diagonal(np.sqrt(np.divide(1, abs(1-hat), where=(1-hat)!=0)))
        total_length = X._nan_triangle().shape[0]
        reshaped_hat = np.reshape(hat[:total_length],(1, total_length))
        indices = np.nansum(X._nan_triangle(), axis=0).cumsum().astype(int)
        for num, item in enumerate(indices[:-1]):
            col_length = int(indices[num+1]-indices[num])
            col = np.reshape(hat[int(indices[num]):int(indices[num+1])],(1,col_length))
            nans = np.repeat(np.expand_dims(np.array([np.nan]),0),total_length-col_length, axis=1)
            col = np.concatenate((col, nans), axis=1)
            reshaped_hat = np.concatenate((reshaped_hat,col),axis=0)
        return reshaped_hat.T

    def _get_hetero_adjustment(self):
        pass

    def _drop(self, X):
        drop = [self.drop] if type(self.drop) is not list else self.drop
        arr = X._nan_triangle()
        for item in drop:
            arr[np.where(X.origin == item[0])[0][0],
                np.where(X.development == item[1])[0][0]] = 0
        return arr

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = copy.copy(X)
        X_new = self.resampled_triangles_
        X_new.scale_ = self.scale_
        return X_new



from chainladder.development.base import DevelopmentBase
import copy
import numpy as np


class DevelopmentConstant(DevelopmentBase):
    """ A Estimator that allows for including of extneral patterns into a
    Development style model.  Currently, this only supports single triangles.
    When this estimator is fit against a triangle, only the grain of the
    existing triangle is retained.


    Parameters
    ----------
    patterns : dict,  (default={})
        A dictionary key/value representation of age(in months)/value 
    style : string, optional (default='ldf')
        type of averaging to use for ldf average calculation.  Options include
        'volume', 'simple', and 'regression'
   

    Attributes
    ----------
    ldf_ : Triangle
        The estimated loss development patterns
    cdf_ : Triangle
        The estimated cumulative development patterns

    """
    def __init__(self, patterns={}, style='ldf'):
        self.patterns = patterns
        self.style = style

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the munich adjustment will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        obj = copy.copy(X)
        obj.values = np.ones(X.shape)[..., :-1]
        ldf = np.array([self.patterns[item] for item in obj.ddims[:-1]])
        if self.style == 'cdf':
            ldf = np.append(ldf[:-1]/ldf[1:], ldf[-1])
        ldf = ldf[np.newaxis, np.newaxis, np.newaxis, ...]
        obj.values = obj.values * ldf
        obj.ddims = X.link_ratio.ddims
        obj.valuation = obj._valuation_triangle(obj.ddims)
        obj.nan_override = True
        obj._set_slicers()

        self.ldf_ = obj
        self.cdf_ = self._get_cdf(self)
        #self.sigma_ = self._param_property(X, params, 1)
        #self.std_err_ = self._param_property(X, params, 2)
        return self

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = copy.copy(X)
        triangles = ['cdf_', 'ldf_']
        for item in triangles:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        return X_new
